from subprocess import CompletedProcess
from unittest import mock

import responses
from flask import g

from dm.domain.entities import Server, ActionTemplate, ActionType, Orchestration, StepExecution, Step, User, \
    OrchExecution
from dm.domain.entities.route import Route
from dm.utils.var_context import VarContext
from dm.web import create_app, db
from dm.web.async_functions import deploy_orchestration
from tests.helpers import TestCaseLockBypass, set_callbacks


class TestDeployOrchestration(TestCaseLockBypass):

    def setUp(self) -> None:
        self.maxDiff = None
        self.app = create_app('test')
        self.app2 = create_app('test')

        self.client = self.app.test_client()
        for app in [self.app, self.app2]:
            with app.app_context():
                db.create_all()
                User.set_initial()
                at1 = ActionTemplate(id='aaaaaaaa-1234-5678-1234-aaaaaaaa0001', name='create dir', version=1,
                                     action_type=ActionType.SHELL, code='useradd {{user}}; mkdir {{dir}}',
                                     parameters={}, expected_stdout='',
                                     expected_rc=0, system_kwargs={})
                at2 = ActionTemplate(id='aaaaaaaa-1234-5678-1234-aaaaaaaa0002', name='rm dir', version=1,
                                     action_type=ActionType.SHELL, code='rmuser {{user}}',
                                     parameters={}, expected_stdout='',
                                     expected_rc=0, system_kwargs={})
                at3 = ActionTemplate(id='aaaaaaaa-1234-5678-1234-aaaaaaaa0003', name='untar', version=1,
                                     action_type=ActionType.SHELL, code='tar -xf {{dir}}',
                                     parameters={}, expected_stdout='',
                                     expected_rc=0, system_kwargs={})
                at4 = ActionTemplate(id='aaaaaaaa-1234-5678-1234-aaaaaaaa0004', name='install tibero',
                                     version=1,
                                     action_type=ActionType.SHELL, code='{{home}}/install_tibero.sh',
                                     parameters={}, expected_stdout='',
                                     expected_rc=0, system_kwargs={})

                o = Orchestration('Test Orchestration', 1, 'description',
                                  id='bbbbbbbb-1234-5678-1234-bbbbbbbb0001')

                me = Server('me', port=5000, me=app == self.app, id='cccccccc-1234-5678-1234-cccccccc0001')
                remote = Server('remote', port=5000, me=app == self.app2,
                                id='cccccccc-1234-5678-1234-cccccccc0002')
                if app == self.app:
                    r = Route(remote, cost=0)
                else:
                    r = Route(me, cost=0)
                db.session.add_all([me, remote, o, r])

                # Orch diagram
                # f   f
                # 1-->3-->u4-->u5-->u6
                #  \/  \    \_____/
                #  /\   \
                # 9-->u2 7-->u8
                # b

                s1 = o.add_step(id='dddddddd-1234-5678-1234-dddddddd0001', undo=False, action_template=at1,
                                parents=[], target=['frontend'])
                s2 = o.add_step(id='dddddddd-1234-5678-1234-dddddddd0002', undo=True, action_template=at2,
                                parents=[s1], target=[])
                s3 = o.add_step(id='dddddddd-1234-5678-1234-dddddddd0003', undo=False, action_template=at3,
                                parents=[s1], target=['frontend'])
                s4 = o.add_step(id='dddddddd-1234-5678-1234-dddddddd0004', undo=True, action_template=at2,
                                parents=[s3], target=[])
                s5 = o.add_step(id='dddddddd-1234-5678-1234-dddddddd0005', undo=True, action_template=at2,
                                parents=[s4], target=[])
                s6 = o.add_step(id='dddddddd-1234-5678-1234-dddddddd0006', undo=True, action_template=at2,
                                parents=[s4, s5], target=[])
                s7 = o.add_step(id='dddddddd-1234-5678-1234-dddddddd0007', undo=False, action_template=at4,
                                parents=[s3], target=[])
                s8 = o.add_step(id='dddddddd-1234-5678-1234-dddddddd0008', undo=True, action_template=at2,
                                parents=[s7], target=[])
                s9 = o.add_step(id='dddddddd-1234-5678-1234-dddddddd0009', undo=False, action_template=at1,
                                children=[s2, s3], target=['backend'])
                db.session.commit()
        self.vs = VarContext(globals=dict(executor_id='00000000-0000-0000-0000-000000000001'),
                             defaults={'user': 'joan', 'dir': '/opt/dimensigon',
                                       'home': '{{dir}}'}, )
        set_callbacks([('me', self.app.test_client()), ('remote', self.app2.test_client())])

    def tearDown(self) -> None:
        for app in [self.app, self.app2]:
            with app.app_context():
                db.session.remove()
                db.drop_all()

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr='')]

        with self.app.app_context():
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            g.server = me
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')

            deploy_orchestration(o.id, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]})

            self.assertEqual(1, OrchExecution.query.count())
            oe = OrchExecution.query.one()
            orch_execution_id = oe.id
            self.assertTrue(oe.success)
            self.assertEqual(o, oe.orchestration)
            self.assertDictEqual(
                {'all': [str(me.id), str(remote.id)], 'frontend': [str(me.id)], 'backend': [str(remote.id)]}, oe.target)
            self.assertDictEqual(dict(self.vs), oe.params)
            self.assertEqual(User.get_by_user('root'), oe.executor)
            self.assertIsNone(oe.service)
            self.assertTrue(oe.success)
            self.assertIsNone(oe.undo_success)

            self.assertEqual(5, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

        with self.app2.app_context():
            self.assertEqual(1, OrchExecution.query.count())
            oe = OrchExecution.query.one()
            self.assertEqual(orch_execution_id, oe.id)
            self.assertEqual(o.id, oe.orchestration.id)
            self.assertDictEqual(
                {'all': [str(me.id), str(remote.id)], 'frontend': [str(me.id)], 'backend': [str(remote.id)]}, oe.target)
            self.assertDictEqual(dict(self.vs), oe.params)
            self.assertEqual(User.get_by_user('root'), oe.executor)
            self.assertIsNone(oe.service)
            # self.assertTrue(oe.success)
            # self.assertIsNone(oe.undo_success)

            self.assertEqual(2, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').one()
            self.assertTrue(e.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_stop_on_error(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr='')]

        with self.app.app_context():
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')

            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]})

            self.assertEqual(1, OrchExecution.query.count())
            oe = OrchExecution.query.one()
            self.assertFalse(oe.success)
            self.assertTrue(oe.undo_success)

            self.assertEqual(8, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').one()
            self.assertFalse(e.success)

            ue4 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0004').one()
            self.assertTrue(ue4.success)

            ue5 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0005').one()
            self.assertTrue(ue5.success)
            self.assertGreater(ue5.start_time, ue4.start_time)

            ue6 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0006').one()
            self.assertTrue(ue6.success)
            self.assertGreater(ue6.start_time, ue5.start_time)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(ue.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(ue.success)

        with self.app2.app_context():
            self.assertEqual(2, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').one()
            self.assertTrue(ue.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_stop_on_error_false(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr='')
                                ]

        with self.app.app_context():
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')
            o.stop_on_error = False
            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]})

            self.assertEqual(12, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').one()
            self.assertFalse(e.success)

            ue4 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0004').one()
            self.assertTrue(ue4.success)

            ue5 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0005').one()
            self.assertTrue(ue5.success)
            self.assertGreater(ue5.start_time, ue4.start_time)

            ue6 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0006').one()
            self.assertTrue(ue6.success)
            self.assertGreater(ue6.start_time, ue5.start_time)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(ue.success)

        with self.app2.app_context():
            self.assertEqual(4, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').one()
            self.assertTrue(ue.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_step_stop_on_error_false(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr='')
                                ]

        with self.app.app_context():
            s = Step.query.get('dddddddd-1234-5678-1234-dddddddd0001')
            s.step_stop_on_error = False
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')
            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]})

            self.assertEqual(12, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertFalse(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(e.success)

            ue81 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(ue81.success)

            ue4 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0004').one()
            self.assertTrue(ue4.success)

            ue5 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0005').one()
            self.assertTrue(ue5.success)
            self.assertGreater(ue5.start_time, ue4.start_time)

            ue6 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0006').one()
            self.assertTrue(ue6.success)
            self.assertGreater(ue6.start_time, ue5.start_time)

            ue21 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(ue21.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(ue.success)

        with self.app2.app_context():
            self.assertEqual(4, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').one()
            self.assertTrue(ue.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_step_stop_on_error_true(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr='')
                                ]

        with self.app.app_context():
            s = Step.query.get('dddddddd-1234-5678-1234-dddddddd0001')
            s.step_stop_on_error = True
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')
            o.stop_on_error = False
            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]})

            self.assertEqual(4, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertFalse(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(ue.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(ue.success)

        with self.app2.app_context():
            self.assertEqual(2, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').one()
            self.assertTrue(ue.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_undo_on_error_false(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr='')]

        with self.app.app_context():
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')
            o.undo_on_error = False
            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]})

            self.assertEqual(5, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertFalse(e.success)

        with self.app2.app_context():
            self.assertEqual(2, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').one()
            self.assertFalse(e.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_step_undo_on_error_false(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''), ]

        with self.app.app_context():
            s = Step.query.get('dddddddd-1234-5678-1234-dddddddd0007')
            s.step_undo_on_error = False
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')
            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]},
                                 )

            self.assertEqual(11, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(ue.success)

            ue4 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0004').one()
            self.assertTrue(ue4.success)

            ue5 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0005').one()
            self.assertTrue(ue5.success)
            self.assertGreater(ue5.start_time, ue4.start_time)

            ue6 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0006').one()
            self.assertTrue(ue6.success)
            self.assertGreater(ue6.start_time, ue5.start_time)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(ue.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertFalse(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(ue.success)

        with self.app2.app_context():
            self.assertEqual(3, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').one()
            self.assertFalse(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').one()
            self.assertTrue(ue.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_stop_undo_on_error_true(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr='')]

        with self.app.app_context():
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')
            o.stop_undo_on_error = True
            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]},
                                 )

            self.assertEqual(5, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').one()
            self.assertFalse(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0004').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0005').one()
            self.assertFalse(ue.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

        with self.app2.app_context():
            self.assertEqual(1, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_step_stop_undo_on_error_true(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr='')]

        with self.app.app_context():
            s = Step.query.get('dddddddd-1234-5678-1234-dddddddd0003')
            s.step_stop_undo_on_error = True
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')
            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]},
                                 )

            self.assertEqual(5, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').one()
            self.assertFalse(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0004').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0005').one()
            self.assertFalse(ue.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

        with self.app2.app_context():
            self.assertEqual(1, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_undo_step_stop_on_error_false(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr='')]

        with self.app.app_context():
            s = Step.query.get('dddddddd-1234-5678-1234-dddddddd0005')
            s.step_stop_on_error = False
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')
            o.stop_undo_on_error = True
            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]},
                                 )

            self.assertEqual(6, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').one()
            self.assertFalse(e.success)

            ue4 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0004').one()
            self.assertTrue(ue4.success)

            ue5 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0005').one()
            self.assertFalse(ue5.success)
            self.assertGreater(ue5.start_time, ue4.start_time)

            ue6 = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0006').one()
            self.assertTrue(ue6.success)
            self.assertGreater(ue6.start_time, ue5.start_time)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)


        with self.app2.app_context():
            self.assertEqual(1, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

    @mock.patch('dm.use_cases.operations.subprocess.run')
    @responses.activate
    def test_deploy_orchestration_stop_undo_on_error_false(self, mock_run):
        mock_run.side_effect = [CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=1, stdout='err', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr=''),
                                CompletedProcess(args=(), returncode=0, stdout='', stderr='')
                                ]

        with self.app.app_context():
            me = Server.query.get('cccccccc-1234-5678-1234-cccccccc0001')
            remote = Server.query.get('cccccccc-1234-5678-1234-cccccccc0002')
            o = Orchestration.query.get('bbbbbbbb-1234-5678-1234-bbbbbbbb0001')
            o.stop_undo_on_error = False
            deploy_orchestration(o, var_context=self.vs,
                                 hosts={'all': [me.id, remote.id], 'frontend': [me.id], 'backend': [remote.id]},
                                 )

            self.assertEqual(12, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0001').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0003').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertFalse(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0004').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0005').one()
            self.assertFalse(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0006').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0001').one()
            self.assertTrue(ue.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertTrue(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').filter_by(
                server_id='cccccccc-1234-5678-1234-cccccccc0002').one()
            self.assertFalse(ue.success)

        with self.app2.app_context():
            self.assertEqual(4, StepExecution.query.count())

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0009').one()
            self.assertTrue(e.success)

            e = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0007').one()
            self.assertTrue(e.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0008').one()
            self.assertFalse(ue.success)

            ue = StepExecution.query.filter_by(step_id='dddddddd-1234-5678-1234-dddddddd0002').one()
            self.assertTrue(ue.success)


import threading
import threading
import typing as t
from abc import ABC

from typing_extensions import Protocol

from dm.domain.exceptions import StateError, ApplicantError, StateAlreadyInPreventingLock, StateAlreadyInLock, \
    StateAlreadyInUnlock, StateTransitionError, PriorityError
from dm.utils.typos import Priority

OE = t.Optional[t.Union[StateError, ApplicantError]]


class Comparable(Protocol):
    """type for objects that support < operator"""

    def __lt__(self, other) -> bool:
        ...


class State(ABC):

    def preventing_lock(self, lock: 'Locker') -> OE:
        return StateAlreadyInPreventingLock()

    def lock(self, lock: 'Locker') -> OE:
        return StateAlreadyInLock()

    def unlock(self, lock: 'Locker') -> OE:
        return StateAlreadyInUnlock()

    def __str__(self):
        return self.__class__.__name__


class UnlockState(State):

    def lock(self, lock: 'Locker') -> OE:
        return StateTransitionError('UNLOCK', 'LOCK')

    def preventing_lock(self, lock: 'Locker') -> OE:
        lock._state = PreventingLockState(lock.unlock, kwargs={'applicant': lock.applicant})
        return


class PreventingLockState(State):

    def __init__(self, func: t.Callable, args: t.Tuple = None, kwargs: t.Mapping[str, t.Any] = None):
        def safe_unlock(func_, *args_, **kwargs_):
            try:
                func_(*args_, **kwargs_)
            except StateError:
                pass

        args = args if args is not None else []
        kwargs = kwargs if kwargs is not None else {}
        self.timer = threading.Timer(interval=Locker.TIMEOUT, function=safe_unlock, args=(func, *args), kwargs=kwargs)
        self.timer.start()

    def lock(self, lock: 'Locker') -> OE:
        self.timer.cancel()
        lock._state = LockState()

    def unlock(self, lock: 'Locker') -> OE:
        self.timer.cancel()
        lock._state = UnlockState()


class LockState(State):

    def preventing_lock(self, lock: 'Locker') -> OE:
        return StateTransitionError('LOCK', 'PREVENTING_LOCK')

    def unlock(self, lock: 'Locker') -> OE:
        lock._state = UnlockState()
        return


# def uid_to_class(uid):
#     for name, cls in inspect.getmembers(sys.modules[__name__], lambda c: inspect.isclass(c) and issubclass(c, State)):
#         if getattr(cls, 'id', None) == uid:
#             return cls
#

class Locker:
    """
    class that holds the state of the lock
    """
    TIMEOUT: t.ClassVar = 90

    def __init__(self):
        self._state: State = UnlockState()
        self._applicant: t.Optional[Comparable] = None  # holds the owner identifier who requested the lock
        self._mutex = threading.Lock()

    def set_timeout(self, timeout: float):
        self.__class__.TIMEOUT = timeout
        return self

    @staticmethod
    def _raise_if_error(msg):
        if isinstance(msg, StateError):
            raise msg

    @property
    def state(self):
        return self._state

    @property
    def applicant(self) -> t.Optional[Comparable]:
        return self._applicant

    def preventing_lock(self, applicant: t.Any):
        with self._mutex:
            if self.applicant is not applicant and self.applicant is not None:
                raise ApplicantError()
            self._applicant = applicant
            msg = self._state.preventing_lock(self)
            self._raise_if_error(msg)

    def lock(self, applicant: t.Any):
        with self._mutex:
            if self.applicant is not applicant and self.applicant is not None:
                raise ApplicantError()
            msg = self._state.lock(self)
            self._raise_if_error(msg)

    def unlock(self, applicant: t.Any):
        with self._mutex:
            if self.applicant is not applicant and self.applicant is not None:
                raise ApplicantError()
            msg = self._state.unlock(self)
            self._raise_if_error(msg)
            self._reset_locker()

    def _reset_locker(self):
        self._applicant = None

    def stop_timer(self):
        if isinstance(self._state, PreventingLockState):
            self._state.timer.cancel()

    def __str__(self):
        return f'{self._state}'


class PriorityLocker:
    """
    Priority Locker class that prevents locking a locker if a more prior locker is trying to be locked

    """

    def __init__(self, priority: Comparable, persistent=False, uid=None):
        """

        Parameters
        ----------
        priority:
            locker priority. Could be any hashable instance that implements __lt__ method. 1 is higher priority than 2
        """
        self.priority = priority
        self._locker = Locker()

    def set_timeout(self, timeout: int):
        self._locker.set_timeout(timeout=timeout)

    @property
    def applicant(self):
        return self._locker.applicant

    @property
    def state(self):
        return self._locker.state

    def preventing_lock(self, lockers: t.Dict[Priority, 'PriorityLocker'], applicant: t.Any):
        # check if higher lockers are locked or in preventing lock
        cond = any(map(lambda s: isinstance(s, (PreventingLockState, LockState)),
                       [locker.state for priority, locker in lockers.items() if priority < self.priority]))
        if not cond:
            self._locker.preventing_lock(applicant=applicant)
        else:
            raise PriorityError()

    def lock(self, applicant: t.Any):
        self._locker.lock(applicant=applicant)

    def unlock(self, applicant: t.Any):
        self._locker.unlock(applicant=applicant)

    def stop_timer(self):
        self._locker.stop_timer()



from unittest import TestCase

from flask_jwt_extended import create_access_token

from dm import defaults
from dm.domain.entities import ActionType, Step, ActionTemplate, Orchestration
from dm.utils.helpers import get_now
from dm.web import create_app, db, errors


class TestStep(TestCase):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        db.create_all()

        self.at1 = ActionTemplate(id='11111111-2222-3333-4444-555555550001', name='action1', version=1,
                                  action_type=ActionType.SHELL, code='code to run', parameters={'param1': 'test'},
                                  expected_stdout='expected output', expected_stderr='stderr', expected_rc=0,
                                  system_kwargs={})

        self.at2 = ActionTemplate(id='11111111-2222-3333-4444-555555550002', name='action2', version=1,
                                  action_type=ActionType.SHELL, code='code to run', parameters={'param1': 'test'},
                                  expected_stdout='expected output', expected_stderr='stderr', expected_rc=0,
                                  system_kwargs={})

        self.o = Orchestration(id='11111111-2222-3333-4444-666666660001', name='orchestration_name',
                               version=1, description='desc')
        # db.session.add_all([self.at1, self.at2, self.o])

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_equality(self):
        created = get_now()
        s1 = Step(orchestration=self.o, undo=True, stop_on_error=False, action_template=self.at1,
                  expected_stdout='expected', expected_stderr='stderr',
                  expected_rc=0, system_kwargs={'timeout': 180}, parameters={'param2': 2}, created_on=created)
        s2 = Step(orchestration=self.o, undo=True, stop_on_error=False, action_template=self.at2,
                  expected_stdout='expected', expected_stderr='stderr',
                  expected_rc=0, system_kwargs={'timeout': 180}, parameters={'param2': 2}, created_on=created)

        self.assertTrue(s1.eq_imp(s2))

        s2.step_parameters.update({'param2': 3})
        self.assertFalse(s1.eq_imp(s2))

        self.at2.parameters['param1'] = 'test2'
        s2.step_parameters.update({'param2': 2, 'param1': 'test'})
        self.at1.parameters.update({'param2': 2})
        self.assertTrue(s1.eq_imp(s2))

    def test_target(self):
        s = Step(None, True, self.at1)
        self.assertEqual(None, s.target)

        with self.assertRaises(ValueError):
            s = Step(None, True, self.at1, target='one')

        s = Step(None, False, self.at1, target='one')
        self.assertEqual(['one'], s.target)

        s = Step(None, False, self.at1)
        self.assertEqual(['all'], s.target)

        s = Step(None, False, self.at1, target=[])
        self.assertEqual(['all'], s.target)

        s = Step(None, False, self.at1, target=['one', 'two'])
        self.assertEqual(['one', 'two'], s.target)

    def test_user_parameters(self):
        at = ActionTemplate(id='11111111-2222-3333-4444-555555550001', name='action1', version=1,
                             action_type=ActionType.SHELL, code='{{param1}}, {{ param2}}, {{param3}}',
                             parameters={'param1': 'param 2 is {{param2}}'},
                             expected_stdout='expected output', expected_stderr='stderr', expected_rc=0,
                             system_kwargs={})

        s = Step(None, undo=False, action_template=at, parameters={'param3': 'value3'})

        self.assertCountEqual(['param2'], s.user_parameters)

    def test_to_from_json(self):
        created = get_now()
        s2 = Step(orchestration=self.o, undo=True,
                  stop_on_error=False,
                  action_template=self.at2,
                  expected_stdout='expected',
                  expected_stderr='stderr',
                  parameters={'param2': 2},
                  id='11111111-2222-3333-4444-111111110002',
                  created_on=created)
        s1 = Step(orchestration=self.o, undo=True,
                  stop_on_error=False,
                  action_template=self.at1,
                  expected_stdout='expected',
                  expected_stderr='stderr',
                  expected_rc=0,
                  system_kwargs={'timeout': 180},
                  parameters={'param2': 2},
                  children_steps=[s2],
                  id='11111111-2222-3333-4444-111111110001',
                  created_on=created)

        s1_json = s1.to_json()
        s2_json = s2.to_json()
        self.assertDictEqual(
            dict(id='11111111-2222-3333-4444-111111110001',
                 orchestration_id='11111111-2222-3333-4444-666666660001',
                 undo=True, stop_on_error=False,
                 undo_on_error=None,
                 stop_undo_on_error=None,
                 action_template_id=str(self.at1.id),
                 error_on_fetch=True,
                 regexp_fetch=None,
                 expected_stdout='expected',
                 expected_stderr='stderr',
                 parent_step_ids=[],
                 expected_rc=0,
                 system_kwargs={'timeout': 180},
                 parameters={'param2': 2},
                 created_on=created.strftime(defaults.DATETIME_FORMAT)), s1_json)
        self.assertDictEqual(
            dict(id='11111111-2222-3333-4444-111111110002',
                 orchestration_id='11111111-2222-3333-4444-666666660001',
                 undo=True, stop_on_error=False,
                 undo_on_error=None,
                 stop_undo_on_error=None,
                 action_template_id=str(self.at2.id),
                 error_on_fetch=True,
                 regexp_fetch=None,
                 expected_stdout='expected',
                 expected_stderr='stderr',
                 parent_step_ids=['11111111-2222-3333-4444-111111110001'],
                 expected_rc=None,
                 system_kwargs={},
                 parameters={'param2': 2},
                 created_on=created.strftime(defaults.DATETIME_FORMAT)), s2_json)

        with self.assertRaises(errors.EntityNotFound):
            Step.from_json(s1_json)

        db.session.add(s1)

        smashed_s1 = Step.from_json(s1_json)
        self.assertEqual(s1, smashed_s1)

        s1_json['parent_step_ids'].append('11111111-2222-3333-4444-111111110003')
        with self.assertRaises(errors.EntityNotFound):
            smashed_s1 = Step.from_json(s1_json)



import base64
import os
import re

from flask import request, current_app
from flask_jwt_extended import jwt_required
from flask_restful import Resource
from sqlalchemy import or_

import dm.defaults as d
from dm.domain.entities import Transfer, TransferStatus, Software
from dm.domain.entities.transfer import Status
from dm.utils.helpers import md5, get_now
from dm.web import db, errors
from dm.web.decorators import securizer, forward_or_dispatch, validate_schema
from dm.web.helpers import filter_query
from dm.web.json_schemas import transfers_post, transfer_post, transfer_patch


class TransferList(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self):
        query = filter_query(Transfer, request.args)
        return [t.to_json() for t in query.all()]

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(transfers_post)
    def post(self):
        # validation
        json_data = request.get_json()
        soft = None
        if 'software_id' in json_data:
            soft = Software.query.get_or_404(json_data['software_id'])
            pending = Transfer.query.filter_by(software=soft, dest_path=json_data['dest_path']).filter(
                or_(Transfer.status == TransferStatus.WAITING_CHUNKS,
                    Transfer.status == TransferStatus.IN_PROGRESS)).all()

            if pending and not json_data.get('cancel_pending', False):
                raise errors.TransferSoftwareAlreadyOpen(str(soft.id))
            elif pending and json_data['cancel_pending']:
                for trans in pending:
                    trans.status = TransferStatus.CANCELED
        elif 'filename' not in json_data:
            return {'error': 'Neither filename nor software specified for the transfer'}, 404
        elif 'filename' in json_data and ('size' not in json_data or 'checksum' not in json_data):
            return {'error': 'size and checksum must be specified when a filename set'}, 404
        else:
            pending = Transfer.query.filter_by(_filename=json_data['filename'],
                                               dest_path=json_data['dest_path']).filter(
                or_(Transfer.status == TransferStatus.WAITING_CHUNKS,
                    Transfer.status == TransferStatus.IN_PROGRESS)).all()

            if pending and not json_data.get('cancel_pending', False):
                raise errors.TransferFileAlreadyOpen(os.path.join(json_data['dest_path'], json_data['filename']))
            elif pending and json_data.get('cancel_pending', False):
                for trans in pending:
                    trans.status = TransferStatus.CANCELED

        dest_path = json_data.get('dest_path', current_app.config['SOFTWARE_REPO'])
        file = os.path.join(dest_path, soft.filename if soft else json_data['filename'])

        if os.path.exists(file):
            if not json_data.get('force', False):
                raise errors.TransferFileAlreadyExists(file)
            else:
                try:
                    os.remove(file)
                except FileNotFoundError:
                    pass
                except Exception as e:
                    msg = f"Unable to remove {file}"
                    current_app.logger.exception(msg)
                    return {'error': msg + f": {e}"}, 500

        # remove chunk files if exist
        for dirpath, dirnames, filenames in os.walk(dest_path):
            for f in filenames:
                if re.search(rf"^{os.path.basename(file)}_chunk\.(\d+)$", f):
                    current_app.logger.debug(f'removing chunk file {os.path.join(dest_path, f)}')
                    try:
                        os.remove(os.path.join(dirpath, f))
                    except:
                        pass
            break

        if soft:
            t = Transfer(software=soft, dest_path=dest_path,
                         num_chunks=json_data['num_chunks'])
        else:
            t = Transfer(software=json_data['filename'], dest_path=dest_path, num_chunks=json_data['num_chunks'],
                         size=json_data['size'], checksum=json_data['checksum'])

        try:
            os.makedirs(t.dest_path, exist_ok=True)
        except Exception as e:
            msg = f"Unable to create dest path {t.dest_path}"
            current_app.logger.exception(msg)
            return {'error': msg + f": {e}"}, 500

        db.session.add(t)
        db.session.commit()
        return {'id': str(t.id)}, 202


CHUNK_READ_BUFFER = d.CHUNK_SIZE


class TransferResource(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, transfer_id):
        return Transfer.query.get_or_404(transfer_id).to_json()

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(transfer_patch)
    def patch(self, transfer_id):
        data = request.get_json()
        trans: Transfer = Transfer.query.get_or_404(transfer_id)
        trans.status = Status[data.get('status')]
        db.session.commit()
        return {'transfer_id': transfer_id, 'status': str(trans.status)}, 200

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(transfer_post)
    def post(self, transfer_id):
        """Generates the chunk into disk"""
        data = request.get_json()
        trans: Transfer = Transfer.query.get_or_404(transfer_id)
        if trans.status == TransferStatus.WAITING_CHUNKS:
            trans.started_on = get_now()
            trans.status = TransferStatus.IN_PROGRESS
            db.session.commit()
        elif trans.status != TransferStatus.IN_PROGRESS:
            raise errors.TransferNotInValidState(transfer_id, trans.status.name)

        chunk = data.get('content')
        chunk_id = data.get('chunk')
        if trans.num_chunks == 1:
            file = os.path.join(trans.dest_path, f'{trans.filename}')
        else:
            file = os.path.join(trans.dest_path, f'{trans.filename}_chunk.{chunk_id}')
        with open(file, 'wb') as fd:
            raw = base64.b64decode(chunk.encode('ascii'))
            fd.write(raw)
        if trans.num_chunks == 1:
            msg = f"File {trans.filename} from transfer {transfer_id} generated successfully"
            trans.status = TransferStatus.COMPLETED
            db.session.commit()
        else:
            msg = f"Chunk {chunk_id} from transfer {transfer_id} generated successfully"

        current_app.logger.debug(msg)
        return {'message': msg}, 201

    @forward_or_dispatch
    @jwt_required
    @securizer
    def put(self, transfer_id):
        """ends the transfer creating the file"""
        trans: Transfer = Transfer.query.get_or_404(transfer_id)
        if trans.status == TransferStatus.COMPLETED:
            return {'error': 'Transfer has already completed'}, 410
        elif trans.status == TransferStatus.WAITING_CHUNKS:
            return {'error': 'Transfer still waiting for chunks'}, 406
        current_app.logger.debug(
            f"Generating file {os.path.join(trans.dest_path, trans.filename)} from transfer {trans.id}")
        chunk_pattern = re.compile(rf"^{trans.filename}_chunk\.(\d+)$")
        file = os.path.join(trans.dest_path, trans.filename)

        try:
            files, chunks_ids = zip(*sorted(
                [(f, int(chunk_pattern.match(f).groups()[0])) for f in os.listdir(trans.dest_path) if
                 os.path.isfile(os.path.join(trans.dest_path, f)) and chunk_pattern.match(f)],
                key=lambda x: x[1]))
        except:
            files, chunks_ids = [], []

        if len(files) != trans.num_chunks or sum(chunks_ids) != (trans.num_chunks - 1) * trans.num_chunks / 2:
            if len(files) == 0:
                msg = f"Any chunk found on {trans.dest_path}"
            else:
                msg = f"Not enough chunks to generate the file"
            current_app.logger.error(msg)
            return {"error": msg}, 404
        with open(file, 'wb') as outfile:
            for fname in files:
                f = os.path.join(trans.dest_path, fname)
                with open(f, 'rb') as infile:
                    while True:
                        c = infile.read(CHUNK_READ_BUFFER)
                        if not c:
                            break
                        outfile.write(c)
                try:
                    os.remove(f)
                except Exception as e:
                    current_app.logger.warning(f"Unable to remove chunk file {f}. Exception: {e}")
        # check final file length and checksum
        if os.path.getsize(file) != trans.size:
            trans.status = TransferStatus.SIZE_ERROR
            db.session.commit()
            # os.remove(file)
            msg = f"Error on transfer '{transfer_id}': Final file size does not match expected size"
            current_app.logger.error(msg)
            return {"error": msg}, 404

        if md5(file) != trans.checksum:
            trans.status = TransferStatus.CHECKSUM_ERROR
            db.session.commit()
            # os.remove(file)
            msg = f"Error on transfer '{transfer_id}': Checksum error"
            current_app.logger.error(msg)
            return {"error": msg}, 404

        trans.status = TransferStatus.COMPLETED
        db.session.commit()
        msg = f"File {os.path.join(trans.dest_path, trans.filename)} from transfer {trans.id} recived successfully"
        current_app.logger.debug(msg)
        return {"message": msg}, 201



import hashlib
import os
import time

import responses
from aioresponses import aioresponses
from flask import url_for
from flask_jwt_extended import create_access_token
from pyfakefs.fake_filesystem_unittest import TestCase

from dm.domain.entities import Server, Route, Dimension, Software, SoftwareServerAssociation, Transfer, TransferStatus
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.utils.helpers import generate_dimension, md5
from dm.web import create_app, db
from tests.helpers import set_callbacks, ValidateResponseMixin


class TestSend(TestCase, ValidateResponseMixin):
    def setUp(self) -> None:
        # create the app with common test config
        self.app1 = create_app('test')
        self.app1.config['SERVER_NAME'] = 'node1'
        # self.app1.config['SECURIZER'] = True
        self.client1 = self.app1.test_client()
        self.app2 = create_app('test')
        self.app2.config['SERVER_NAME'] = 'node2'
        # self.app2.config['SECURIZER'] = True
        self.client2 = self.app2.test_client()

        self.source_path = '/software'
        self.filename = 'filename.zip'
        self.content = b'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
        self.size = len(self.content)
        self.checksum = hashlib.md5(self.content).hexdigest()
        self.dest_path = '/dest_repo'

        with self.app1.app_context():
            db.create_all()
            set_initial()
            s = Server.get_current()
            s.gates = []
            s.add_new_gate('node1', 5000)
            soft = Software(name='test_software', version=1, filename=self.filename, size=self.size,
                            checksum=self.checksum)
            ssa = SoftwareServerAssociation(software=soft, server=s, path=self.source_path)
            dim = generate_dimension('dimension')
            dim.current = True
            db.session.add_all([s, dim, soft, ssa])
            db.session.commit()
            self.s1_json = Server.get_current().to_json()
            self.soft_json = soft.to_json()
            self.ssa_json = ssa.to_json()
            self.dim_json = dim.to_json()
            self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))

        with self.app2.app_context():
            db.create_all()
            set_initial()
            s = Server.get_current()
            s.gates = []
            s.add_new_gate('node2', 5000)
            db.session.commit()
            self.s2_json = Server.get_current().to_json()
            s = Server.from_json(self.s1_json)
            s.add_new_gate('node1', 5000)
            Route(s, cost=0)

            db.session.add(s)
            dim = Dimension.from_json(self.dim_json)
            dim.current = True
            db.session.add(dim)

            soft = Software.from_json(self.soft_json)
            db.session.add(soft)

            ssa = SoftwareServerAssociation.from_json(self.ssa_json)
            db.session.add(ssa)
            db.session.commit()

        self.context = self.app1.app_context()
        self.context.push()
        s = Server.from_json(self.s2_json)
        s.add_new_gate('node2', 5000)
        Route(s, cost=0)
        db.session.add(s)

        self.remote = s

        self.setUpPyfakefs()
        self.fs.create_dir(self.source_path)
        self.fs.create_dir(self.dest_path)
        self.fs.create_file(os.path.join(self.source_path, self.filename), contents=self.content)

        db.session.commit()


    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.context.pop()

        with self.app2.app_context():
            db.session.remove()
            db.drop_all()

    @aioresponses()
    @responses.activate
    def test_send(self, m):
        set_callbacks([('node2', self.client2)], m)

        server = Server.query.filter_by(name='node2').one()

        self.assertFalse(os.path.exists(os.path.join(self.dest_path, self.filename)))

        resp = self.client1.post(url_for('api_1_0.send'),
                                 json=dict(software_id=self.soft_json['id'], dest_server_id=server.id,
                                           dest_path=self.dest_path),
                                 headers=self.auth.header)
        self.assertEqual(202, resp.status_code)
        self.assertIn('transfer_id', resp.get_json())

        with self.app2.app_context():
            t = Transfer.query.get(resp.get_json().get('transfer_id'))
            start = time.time()
            while t.status != TransferStatus.COMPLETED and (time.time() - start) < 5:
                time.sleep(0.1)
                db.session.refresh(t)

        self.assertTrue(os.path.exists(os.path.join(self.dest_path, self.filename)))
        self.assertEqual(self.size, os.path.getsize(os.path.join(self.dest_path, self.filename)))
        self.assertEqual(self.checksum, md5(os.path.join(self.dest_path, self.filename)))


import socket
import ssl
import time
from contextlib import closing


def is_open(host: str, port: int, timeout: float = 1.0):
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(timeout)
        if sock.connect_ex((host, port)) == 0:
            return True
        else:
            return False


def is_ssl_open(host: str, port: int, timeout: float = 1.0):
    context = ssl.SSLContext()
    try:
        with socket.create_connection((host, port), timeout=timeout) as sock:
            with context.wrap_socket(sock, server_hostname=host) as ssock:
                return True
    except (ConnectionRefusedError, TimeoutError):
        return False
    except socket.timeout:
        return False


def is_open2(host: str, port: int, timeout: float = 1.0):
    context = ssl.SSLContext()
    try:
        with socket.create_connection((host, port), timeout=timeout) as sock:
            with context.wrap_socket(sock, server_hostname=host) as ssock:
                return True
    except (ConnectionRefusedError, TimeoutError):
        return False
    except socket.timeout:
        return True


def check_host(host: str, port: int, retry=3, delay=2, timeout=1.0):
    ipup = False
    for i in range(retry):
        if is_ssl_open(host, port, timeout):
            ipup = True
            break
        elif is_open(host, port, timeout):
            ipup = True
            break
        time.sleep(delay)
    return ipup



from subprocess import PIPE, CompletedProcess
from unittest import TestCase, mock

import dm.use_cases.operations


class TestNativeCommand(TestCase):
    def test_execute(self):
        nc = dm.use_cases.operations.ShellOperation('mkdir {{dir}}', expected_stdout=None, expected_rc=None,
                                                     system_kwargs=None)
        with mock.patch('dm.utils.subprocess.run', return_value=CompletedProcess(args=(), returncode=0, stdout='stdout',
                                                                        stderr='stderr')) as mocked_run:
            rc = nc.execute(dict({'dir': 'c:\\test'}))
            mocked_run.assert_called_once_with('mkdir c:\\test', stdout=PIPE, stderr=PIPE, shell=True, timeout=300)
            self.assertTrue(rc.success)

        nc = dm.use_cases.operations.ShellOperation('mkdir {{dir}}', expected_stdout=None, expected_rc=None,
                                                     system_kwargs={'timeout': 50})
        with mock.patch('dm.utils.subprocess.run', return_value=CompletedProcess(args=(), returncode=0, stdout='stdout',
                                                                        stderr='stderr')) as mocked_run:
            rc = nc.execute(dict({'dir': 'c:\\test'}))
            mocked_run.assert_called_once_with('mkdir c:\\test', stdout=PIPE, stderr=PIPE, shell=True, timeout=50)
            self.assertTrue(rc.success)

        nc = dm.use_cases.operations.ShellOperation('mkdir {{dir}}', expected_stdout='stdout', expected_rc=None,
                                                     system_kwargs=None)
        with mock.patch('dm.utils.subprocess.run', return_value=CompletedProcess(args=(), returncode=0, stdout='',
                                                                        stderr='stderr')):
            rc = nc.execute(dict({'dir': 'c:\\test'}))
            self.assertFalse(rc.success)

        nc = dm.use_cases.operations.ShellOperation('mkdir {dir}', expected_stdout='stdout', expected_rc=0,
                                                     system_kwargs=None)
        with mock.patch('dm.utils.subprocess.run', return_value=CompletedProcess(args=(), returncode=1, stdout='stdout',
                                                                        stderr='stderr')):
            rc = nc.execute(dict({'dir': 'c:\\test'}))
            self.assertFalse(rc.success)

        nc = dm.use_cases.operations.ShellOperation('mkdir {{dir}}', expected_stdout='stdout', expected_rc=1,
                                                     system_kwargs=None)
        with mock.patch('dm.utils.subprocess.run', return_value=CompletedProcess(args=(), returncode=1, stdout='stdout',
                                                                        stderr='stderr')):
            rc = nc.execute(dict({'dir': 'c:\\test'}))
            self.assertTrue(rc.success)



import argparse
import types
from typing import Text, Optional, TypeVar, NoReturn


class ArgumentParserRaise(argparse.ArgumentParser):

    def exit(self, status: int = ..., message: Optional[Text] = ...) -> NoReturn:
        if message != ...:
            print(message)
        raise SystemExit


_T = TypeVar('_T')


class GuessArgumentParser(ArgumentParserRaise):

    # =====================================
    # Command line argument parsing methods
    # =====================================
    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)
        return args

    def parse_known_args(self, args=None, namespace=None):
        if args is None:
            # args default to the system args
            args = argparse._sys.argv[1:]
        else:
            # make sure that args are mutable
            args = list(args)

        # default Namespace built from parser defaults
        if namespace is None:
            namespace = argparse.Namespace()

        # add any action defaults that aren't present
        for action in self._actions:
            if action.dest is not argparse.SUPPRESS:
                if not hasattr(namespace, action.dest):
                    if action.default is not argparse.SUPPRESS:
                        setattr(namespace, action.dest, action.default)

        # add any parser defaults that aren't present
        for dest in self._defaults:
            if not hasattr(namespace, dest):
                setattr(namespace, dest, self._defaults[dest])

        # parse the arguments and exit if there are any errors
        try:
            namespace, args = self._parse_known_args(args, namespace)
            if hasattr(namespace, argparse._UNRECOGNIZED_ARGS_ATTR):
                args.extend(getattr(namespace, argparse._UNRECOGNIZED_ARGS_ATTR))
                delattr(namespace, argparse._UNRECOGNIZED_ARGS_ATTR)
            return namespace, args
        except argparse.ArgumentError:
            err = argparse._sys.exc_info()[1]
            self.error(str(err))

    def _check_value(self, action, value):
        pass

    def _parse_known_args(self, arg_strings, namespace):

        # map all mutually exclusive arguments to the other arguments
        # they can't occur with
        action_conflicts = {}

        # find all option indices, and determine the arg_string_pattern
        # which has an 'O' if there is an option at an index,
        # an 'A' if there is an argument, or a '-' if there is a '--'
        option_string_indices = {}
        arg_string_pattern_parts = []
        arg_strings_iter = iter(arg_strings)
        for i, arg_string in enumerate(arg_strings_iter):

            # all args after -- are non-options
            if arg_string == '--':
                arg_string_pattern_parts.append('-')
                for arg_string in arg_strings_iter:
                    arg_string_pattern_parts.append('A')

            # otherwise, add the arg to the arg strings
            # and note the index if it was an option
            else:
                option_tuple = self._parse_optional(arg_string)
                if option_tuple is None:
                    pattern = 'A'
                else:
                    option_string_indices[i] = option_tuple
                    pattern = 'O'
                arg_string_pattern_parts.append(pattern)

        # join the pieces together to form the pattern
        arg_strings_pattern = ''.join(arg_string_pattern_parts)

        # converts arg strings to the appropriate and then takes the action
        seen_actions = set()
        seen_non_default_actions = set()

        def take_action(action, argument_strings, option_string=None):
            seen_actions.add(action)
            argument_values = self._get_values(action, argument_strings)

            # error if this argument is not allowed with other previously
            # seen arguments, assuming that actions that use the default
            # value don't really count as "present"
            if argument_values is not action.default:
                seen_non_default_actions.add(action)

            # take the action if we didn't receive a SUPPRESS value
            # (e.g. from a default)
            if argument_values is not argparse.SUPPRESS:
                try:
                    action(self, namespace, argument_values, option_string)
                except:
                    pass

        # function to convert arg_strings into an optional action
        def consume_optional(start_index):

            # get the optional identified at this index
            option_tuple = option_string_indices[start_index]
            action, option_string, explicit_arg = option_tuple

            # identify additional optionals in the same arg string
            # (e.g. -xyz is the same as -x -y -z if no args are required)
            match_argument = self._match_argument
            action_tuples = []
            while True:

                # if we found no optional action, skip it
                if action is None:
                    extras.append(arg_strings[start_index])
                    return start_index + 1

                # if there is an explicit argument, try to match the
                # optional's string arguments to only this
                if explicit_arg is not None:
                    arg_count = match_argument(action, 'A')

                    # if the action is a single-dash option and takes no
                    # arguments, try to parse more single-dash options out
                    # of the tail of the option string
                    chars = self.prefix_chars
                    if arg_count == 0 and option_string[1] not in chars:
                        action_tuples.append((action, [], option_string))
                        char = option_string[0]
                        option_string = char + explicit_arg[0]
                        new_explicit_arg = explicit_arg[1:] or None
                        optionals_map = self._option_string_actions
                        if option_string in optionals_map:
                            action = optionals_map[option_string]
                            explicit_arg = new_explicit_arg

                    # if the action expect exactly one argument, we've
                    # successfully matched the option; exit the loop
                    elif arg_count == 1:
                        stop = start_index + 1
                        args = [explicit_arg]
                        action_tuples.append((action, args, option_string))
                        break

                # if there is no explicit argument, try to match the
                # optional's string arguments with the following strings
                # if successful, exit the loop
                else:
                    start = start_index + 1
                    selected_patterns = arg_strings_pattern[start:]
                    arg_count = match_argument(action, selected_patterns)
                    stop = start + arg_count
                    args = arg_strings[start:stop]
                    action_tuples.append((action, args, option_string))
                    break

            # add the Optional to the list and return the index at which
            # the Optional's string args stopped
            assert action_tuples
            for action, args, option_string in action_tuples:
                take_action(action, args, option_string)
            return stop

        # the list of Positionals left to be parsed; this is modified
        # by consume_positionals()
        positionals = self._get_positional_actions()

        # function to convert arg_strings into positional actions
        def consume_positionals(start_index):
            # match as many Positionals as possible
            match_partial = self._match_arguments_partial
            selected_pattern = arg_strings_pattern[start_index:]
            arg_counts = match_partial(positionals, selected_pattern)

            # slice off the appropriate arg strings for each Positional
            # and add the Positional and its args to the list
            for action, arg_count in zip(positionals, arg_counts):
                args = arg_strings[start_index: start_index + arg_count]
                start_index += arg_count
                take_action(action, args)

            # slice off the Positionals that we just parsed and return the
            # index at which the Positionals' string args stopped
            positionals[:] = positionals[len(arg_counts):]
            return start_index

        # consume Positionals and Optionals alternately, until we have
        # passed the last option string
        extras = []
        start_index = 0
        if option_string_indices:
            max_option_string_index = max(option_string_indices)
        else:
            max_option_string_index = -1
        while start_index <= max_option_string_index:

            # consume any Positionals preceding the next option
            next_option_string_index = min([
                index
                for index in option_string_indices
                if index >= start_index])
            if start_index != next_option_string_index:
                positionals_end_index = consume_positionals(start_index)

                # only try to parse the next optional if we didn't consume
                # the option string during the positionals parsing
                if positionals_end_index > start_index:
                    start_index = positionals_end_index
                    continue
                else:
                    start_index = positionals_end_index

            # if we consumed all the positionals we could and we're not
            # at the index of an option string, there were extra arguments
            if start_index not in option_string_indices:
                strings = arg_strings[start_index:next_option_string_index]
                extras.extend(strings)
                start_index = next_option_string_index

            # consume the next optional and any arguments for it
            start_index = consume_optional(start_index)

        # consume any positionals following the last Optional
        stop_index = consume_positionals(start_index)

        # if we didn't consume all the argument strings, there were extras
        extras.extend(arg_strings[stop_index:])

        # make sure all required actions were present and also convert
        # action defaults which were not given as arguments
        required_actions = []
        for action in self._actions:
            if action not in seen_actions:
                if action.required:
                    required_actions.append(argparse._get_action_name(action))
                else:
                    # Convert action default now instead of doing it before
                    # parsing arguments to avoid calling convert functions
                    # twice (which may fail) if the argument was given, but
                    # only if it was defined already in the namespace
                    if (action.default is not None and
                            isinstance(action.default, str) and
                            hasattr(namespace, action.dest) and
                            action.default is getattr(namespace, action.dest)):
                        setattr(namespace, action.dest,
                                self._get_value(action, action.default))

        # make sure all required groups had one option present
        for group in self._mutually_exclusive_groups:
            if group.required:
                for action in group._group_actions:
                    if action in seen_non_default_actions:
                        break

        # return the updated namespace and the extra arguments
        return namespace, extras


class DictAction(argparse.Action):
    def __init__(self, option_strings, dest, **kwargs):
        super(DictAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        container = getattr(namespace, self.dest, None)
        if container is None:
            setattr(namespace, self.dest, {values[0]: values[1:]})
        else:
            container.update({values[0]: values[1:]})


class ParamAction(argparse.Action):
    def __init__(self, option_strings, dest, **kwargs):
        super(ParamAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if len(values) == 1:

            if '=' not in values[0]:
                raise ValueError(f"Not a valid parameter '{values}'")
            else:
                key, value = values[0].split('=', 1)
        else:
            key = values[0]
            value = ' '.join(values[1:])
        container = getattr(namespace, self.dest, None)
        if container is None:
            setattr(namespace, self.dest, {key: value})
        else:
            container.update({key: value})


def create_parser(data, parser=None) -> ArgumentParserRaise:
    if parser is None:
        parser = ArgumentParserRaise()

    if isinstance(parser, GuessArgumentParser):
        pop_keys = ('argument', 'completer', 'type')
    else:
        pop_keys = ('argument', 'completer')
    if isinstance(data, dict):
        subparsers = parser.add_subparsers()
        for k, v in data.items():
            i_parser = subparsers.add_parser(k)
            if isinstance(v, dict):
                create_parser(v, parser=i_parser)
            elif isinstance(v, list):
                for arg in v:
                    if isinstance(arg, list):
                        group = i_parser.add_mutually_exclusive_group()
                        for a in arg:
                            group.add_argument(a['argument'],
                                               **{kk: vv for kk, vv in a.items() if
                                                  kk not in pop_keys})
                    elif isinstance(arg, types.FunctionType):
                        i_parser.set_defaults(func=arg)
                    else:
                        i_parser.add_argument(arg['argument'],
                                              **{kk: vv for kk, vv in arg.items() if
                                                 kk not in pop_keys})
    else:
        for arg in data:
            if isinstance(arg, list):
                group = parser.add_mutually_exclusive_group()
                for a in arg:
                    group.add_argument(a['argument'],
                                       **{kk: vv for kk, vv in a.items() if kk not in pop_keys})
            elif isinstance(arg, types.FunctionType):
                parser.set_defaults(func=arg)
            else:
                parser.add_argument(arg['argument'],
                                    **{kk: vv for kk, vv in arg.items() if kk not in pop_keys})
    return parser



import datetime as dt
import random
from unittest import TestCase, mock

from flask_jwt_extended import create_access_token

from dm import defaults
from dm.domain.entities import User, Server, Route, Scope
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.use_cases.helpers import get_servers_from_scope
from dm.web import create_app, db


class Test(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.create_all()
        set_initial(user=True)
        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @mock.patch('dm.domain.entities.get_now')
    def test_get_servers_from_scope_more_than_min_quorum(self, mock_get_now):
        mock_get_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)

        servers = []
        for i in range(0, 7):

            s = Server(f'node{i}', port=5000)

            if i == 0:
                r = Route(s, cost=0)
            else:
                r = Route(s, proxy_server=random.choice(servers), cost=i)

            db.session.add_all([s, r])

            servers.append(s)

        mock_get_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
        s62 = Server(f'node72', port=5000)
        Route(s62, proxy_server=random.choice(servers), cost=6)
        db.session.add(s62)

        quorum = get_servers_from_scope(scope=Scope.CATALOG)

        self.assertEqual(8, len(quorum))
        self.assertNotIn(s62, quorum)
        self.assertIn(s, quorum)
        self.assertIn(Server.get_current(), quorum)

    @mock.patch('dm.domain.entities.get_now')
    def test_get_servers_from_scope_less_than_min_quorum(self, mock_get_now):
        mock_get_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)

        servers = []
        for i in range(1, 4):
            s = Server(f'node{i}', port=5000)
            r = Route(s, cost=0)

            db.session.add_all([s, r])

            servers.append(s)

        quorum = get_servers_from_scope(scope=Scope.CATALOG)

        self.assertEqual(4, len(quorum))
        self.assertIn(Server.get_current(), quorum)

    @mock.patch('dm.domain.entities.get_now')
    def test_get_servers_from_scope_more_than_min_quorum_no_cost(self, mock_get_now):
        mock_get_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)

        servers = []
        for i in range(0, 7):

            s = Server(f'node{i}', port=5000)

            if i == 0:
                r = Route(s, cost=0)
            else:
                r = Route(s, cost=0)

            db.session.add_all([s, r])

            servers.append(s)

        mock_get_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
        s62 = Server(f'node72', port=5000)
        Route(s62, proxy_server=random.choice(servers), cost=6)
        db.session.add(s62)

        quorum = get_servers_from_scope(scope=Scope.CATALOG)

        self.assertEqual(defaults.MIN_SERVERS_QUORUM+1, len(quorum))
        self.assertIn(Server.get_current(), quorum)



from collections import OrderedDict
from concurrent.futures import Future

from dm.web.extensions.flask_executor.helpers import InstanceProxy


class FutureCollection:
    """A FutureCollection is an object to store and interact with
    :class:`concurrent.futures.Future` objects. It provides access to all
    attributes and methods of a Future by proxying attribute calls to the
    stored Future object.
    To access the methods of a Future from a FutureCollection instance, include
    a valid ``future_key`` value as the first argument of the method call. To
    access attributes, call them as though they were a method with
    ``future_key`` as the sole argument. If ``future_key`` does not exist, the
    call will always return None. If ``future_key`` does exist but the
    referenced Future does not contain the requested attribute an
    :exc:`AttributeError` will be raised.
    To prevent memory exhaustion a FutureCollection instance can be bounded by
    number of items using the ``max_length`` parameter. As a best practice,
    Futures should be popped once they are ready for use, with the proxied
    attribute form used to determine whether a Future is ready to be used or
    discarded.
    :param max_length: Maximum number of Futures to store. Oldest Futures are
    discarded first.
    """

    def __init__(self, max_length=50):
        self.max_length = max_length
        self._futures = OrderedDict()

    def __contains__(self, future):
        return future in self._futures.values()

    def __len__(self):
        return len(self._futures)

    def __getattr__(self, attr):
        # Call any valid Future method or attribute
        def _future_attr(future_key, *args, **kwargs):
            if future_key not in self._futures:
                return None
            future_attr = getattr(self._futures[future_key], attr)
            if callable(future_attr):
                return future_attr(*args, **kwargs)
            return future_attr

        return _future_attr

    def _check_limits(self):
        if self.max_length is not None:
            while len(self._futures) > self.max_length:
                self._futures.popitem(last=False)

    def add(self, future_key, future):
        """Add a new Future. If ``max_length`` limit was defined for the
        FutureCollection, old Futures may be dropped to respect this limit.
        :param future_key: Key for the Future to be added.
        :param future: Future to be added.
        """
        if future_key in self._futures:
            raise ValueError("future_key {} already exists".format(future_key))
        self._futures[future_key] = future
        self._check_limits()

    def pop(self, future_key):
        """Return a Future and remove it from the collection. Futures that are
        ready to be used should always be popped so they do not continue to
        consume memory.
        Returns ``None`` if the key doesn't exist.
        :param future_key: Key for the Future to be returned.
        """
        return self._futures.pop(future_key, None)


class FutureProxy(InstanceProxy, Future):
    """A FutureProxy is an instance proxy that wraps an instance of
    :class:`concurrent.futures.Future`. Since an executor can't be made to
    return a subclassed Future object, this proxy class is used to override
    instance behaviours whilst providing an agnostic method of accessing
    the original methods and attributes.
    :param future: An instance of :class:`~concurrent.futures.Future` that
                   the proxy will provide access to.
    :param executor: An instance of :class:`flask_executor.Executor` which
                     will be used to provide access to Flask context features.
    """

    def __init__(self, future, executor):
        self._self = future
        self._executor = executor

    def add_done_callback(self, fn):
        fn = self._executor._prepare_fn(fn, force_copy=True)
        return self._self.add_done_callback(fn)


import base64
import concurrent
import datetime
import json
import logging
import pickle
import re
import threading
import time
import typing as t
from abc import ABC, abstractmethod
from collections import ChainMap
from concurrent.futures.process import ProcessPoolExecutor
from functools import partial

from RestrictedPython import compile_restricted, safe_builtins
from flask import current_app, Flask, has_app_context
from flask_jwt_extended import create_access_token
from sqlalchemy.orm import sessionmaker

from dm import defaults
from dm.domain.entities import Server, Orchestration, Step, StepExecution, OrchExecution
from dm.network.auth import HTTPBearerAuth
from dm.use_cases.operations import CompletedProcess, IOperationEncapsulation, create_operation
from dm.utils.dag import DAG
from dm.utils.event_handler import Event
from dm.utils.typos import Id
from dm.utils.var_context import VarContext
from dm.web import db
from dm.web.network import post

# if t.TYPE_CHECKING:


logger = logging.getLogger('dm.deployment')


class ICommand(ABC):

    def __init__(self, id_):
        self._id = id_ or id(self)

    @property
    @abstractmethod
    def success(self) -> bool:
        """
        Property that returns True if the command executed successfully
        Returns
        -------

        """

    @property
    def id(self) -> Id:
        """
        Property that returns the ID of the command
        Returns
        -------

        """
        return self._id

    @property
    @abstractmethod
    def result(self) -> t.Dict[Id, CompletedProcess]:
        """
        Property to get all the results

        Returns
        -------
        dict with all the results
        """

    @abstractmethod
    def invoke(self, timeout=None) -> bool:
        """
        Implement the do process. If the process ends as expected it must set _success to true.

        Returns
        -------
        bool:
            True if the do process ends as expected. False otherwise
        """

    @abstractmethod
    def undo(self, timeout=None) -> t.Optional[bool]:
        """
        Method to implement the undo process. Executed only if success is true

        Returns
        -------
        bool:
            True if the undo process ends up OK. False otherwise
        """

    def __repr__(self):
        return f"{self.__class__.__name__} {self.id}"


class ImplementationCommand(ICommand):

    def __init__(self, implementation: IOperationEncapsulation,
                 var_context: VarContext = None,
                 id_=None,
                 register: 'RegisterStepExecution' = None,
                 pre_process=None,
                 post_process=None,
                 ):
        super().__init__(id_)
        self.implementation = implementation
        self.var_context = var_context or {}
        self.register = register
        self.pre_process = pre_process
        self.post_process = post_process
        self._cp: CompletedProcess = None

    def invoke(self, timeout=None) -> t.Optional[bool]:
        if not self._cp:
            try:
                self.pre_processing(local=dict(vs=self.var_context, cp=self._cp))
            except Exception as e:
                self._cp.pre_post_error = e
                self._cp.success = False
                return self._cp.success

            self._cp = self.implementation.execute(self.var_context, timeout=timeout)

            try:
                self.post_processing(local=dict(vc=self.var_context, cp=self._cp))
            except Exception as e:
                self._cp.pre_post_error = e
                self._cp.success = False

            if self.register:
                self.register.register_step_execution(self)

        return self.success

    @property
    def result(self) -> t.Dict[Id, CompletedProcess]:
        return {self.id: self._cp}

    @property
    def success(self) -> t.Optional[bool]:
        return getattr(self._cp, 'success', None)

    def pre_processing(self, local: dict = None):
        local = local or dict()
        if self.pre_process:
            byte_code = compile_restricted(self.pre_process, '<inline>', 'exec')
            safe_builtins.update(json=json)
            exec(byte_code, {'__builtins__': safe_builtins, '_write_': lambda x: x}, local)

    def post_processing(self, local: dict = None):
        local = local or dict()
        if self.post_process:
            if set(re.findall(r'\(\?P<(\w+)>', self.post_process, flags=re.MULTILINE)):
                fetched = {}
                match = re.search(self.post_process, self._cp.stdout)
                if not match:
                    match = re.search(self.post_process, self._cp.stderr)
                    if match:
                        fetched.update(match.groupdict())
                else:
                    fetched.update(match.groupdict())
                for k, v in fetched.items():
                    self.var_context.set(k, v)
            else:
                byte_code = compile_restricted(self.post_process, '<inline>', 'exec')
                safe_builtins.update(json=json)
                exec(byte_code, {'__builtins__': safe_builtins, '_write_': lambda x: x}, local)


class UndoCommand(ImplementationCommand):

    def __init__(self, *args, stop_on_error: bool = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.stop_on_error = stop_on_error

    def undo(self, timeout=None) -> t.Optional[bool]:
        return True

    def __iter__(self):
        return [self]


class Command(ImplementationCommand):

    def __init__(self, *args,
                 undo_command: t.Union['CompositeCommand', UndoCommand] = None,
                 stop_on_error: bool = None, stop_undo_on_error: bool = None, undo_on_error: bool = True, **kwargs
                 ):
        """
        
        Parameters
        ----------
        implementation:
            operation to perform.
        undo_command:
            undo command
        var_context:
            var_context used to execute the implementation
        undo_on_error:
            sets whether to execute "undo" function when "invoke" terminated incorrectly
        """
        super().__init__(*args, **kwargs)
        self.undo_command = undo_command
        self.stop_on_error = stop_on_error
        self.stop_undo_on_error = stop_undo_on_error
        self.undo_on_error = undo_on_error
        self._cp: t.Optional[CompletedProcess] = None


    @property
    def result(self) -> t.Dict[Id, CompletedProcess]:
        e = {}
        e.update({self.id: self._cp})
        if self.undo_command and self.undo_command.success is not None:
            e.update(self.undo_command.result)
        return e

    def undo(self, timeout=None) -> t.Optional[bool]:
        """
        Executes undo operation

        Returns
        -------
        bool:
            True if all undo commands that run ended up successfully
            False if any undo command ended up badly
            None if undo operation not executed
        """
        if self.undo_command:
            if (self.success is True or (
                    self.success is False and self.undo_on_error)) and self.undo_command.success is None:
                return self.undo_command.invoke(timeout=timeout)
            else:
                return self.undo_command.success
        else:
            return True

    def __iter__(self):
        return [self]


class CompositeCommand(ICommand):

    def __init__(self, dict_tree: t.Dict[ICommand, t.List[ICommand]],
                 stop_on_error: bool, stop_undo_on_error: bool = None,
                 id_=None, executor: concurrent.futures.Executor = None, timeout: t.Union[int, float] = None):
        """

        Parameters
        ----------
        dict_tree:
            dict tree containing command nodes as keys and it's neighbours as a list of commands. See examples for more
            information
        undo_on_error:
            executes the undo process even if the command ended up with error
        stop_on_error:
            executes all the commands regardless if a command succeed or fail
        stop_undo_on_error:
            executes all the undo commands regardless if an undo command succeed or fail
        id_:
            command identifier
        executor:
            async call executor. defaults to concurrent.futures.process.ProcessPoolExecutor
        """
        super().__init__(id_)
        self._dag = DAG().from_dict_of_lists(dict_tree)
        self.stop_on_error = stop_on_error
        self.stop_undo_on_error = stop_undo_on_error
        self.executor = executor or ProcessPoolExecutor(max_workers=4)
        self.timeout = timeout

    @property
    def success(self) -> t.Optional[bool]:
        success_list = [n.success for n in self._dag]
        if success_list == [None] * len(success_list):
            return None
        elif success_list == [True] * len(success_list):
            return True
        else:
            return False

    def invoke(self, timeout=None) -> bool:
        res = []
        stop = False
        level = 1
        timeout = timeout or self.timeout
        start = time.time()
        while level <= self._dag.depth:
            commands = self._dag.get_nodes_at_level(level)
            duration = time.time() - start
            left = max(timeout - duration, 0) if timeout else None
            if len(commands) == 1:
                r = commands[0].invoke(timeout=left)
                res.append(r)
                stop_on_error = commands[0].stop_on_error if commands[0].stop_on_error is not None \
                    else self.stop_on_error
                if r is False and stop_on_error is True:
                    break
            else:
                futures = {}

                for command in commands:
                    futures.update(
                        {command.id: self.executor.submit(command.invoke, timeout=left)})

                stop = None
                while len(futures) > 0 and (left is None or (time.time() - start) < left) and stop is not True:
                    for cmd_id, future in dict(futures).items():
                        r = None
                        try:
                            r = future.result(timeout=0.5)
                        except concurrent.futures.TimeoutError:
                            pass
                        except Exception as e:
                            logger.exception(f"Error while executing step {cmd_id}")
                            del futures[cmd_id]
                            r = False
                        else:
                            del futures[cmd_id]
                        if r is not None:
                            res.append(r)
                            cmd = list(filter(lambda c: c.id == cmd_id, commands))[0]
                            stop_on_error = cmd.stop_on_error if cmd.stop_on_error is not None else self.stop_on_error
                            if r is False and stop_on_error is True:
                                stop = True

                    duration = time.time() - start
                    left = max(timeout - duration, 0) if timeout else None

                if stop is True or (timeout is not None and (time.time() - start) > timeout):
                    break
            level += 1
        return all(res) if len(res) > 0 else None

    def undo(self, timeout=None) -> t.Optional[bool]:
        """
        Executes undo operation

        Returns
        -------
        bool:
            True if all undo commands that run ended up successfully
            False if any undo command ended up badly
            None if undo operation not executed
        """
        if self.stop_undo_on_error is None:
            raise RuntimeError(f'stop_undo_on_error not set for command {self.id}')
        res = []
        level = self._dag.depth
        start = time.time()
        duration = time.time() - start
        left = max(timeout - duration, 0) if timeout else None
        while level > 0 and (left is None or (time.time() - start) < left):
            commands = self._dag.get_nodes_at_level(level)
            duration = time.time() - start
            left = max(timeout - duration, 0) if timeout else None
            if len(commands) == 1:
                r = commands[0].undo(timeout=left)
                if r is not None:
                    res.append(r)
                stop_undo_on_error = commands[0].stop_undo_on_error if commands[0].stop_undo_on_error is not None \
                    else self.stop_undo_on_error
                if r is False and stop_undo_on_error is True:
                    break
            else:
                futures = {}

                for command in commands:
                    futures.update(
                        {command.id: self.executor.submit(command.undo, timeout=left)})

                stop = None
                while len(futures) > 0 and (left is None or (time.time() - start) < left):
                    for cmd_id, future in dict(futures).items():
                        r = None
                        try:
                            r = future.result(timeout=0.5)
                        except concurrent.futures.TimeoutError:
                            pass
                        except Exception as e:
                            logger.exception(f"Error while executing step {cmd_id}")
                            del futures[cmd_id]
                            r = False
                        else:
                            del futures[cmd_id]
                        if r is not None:
                            res.append(r)
                            cmd = list(filter(lambda c: c.id == cmd_id, commands))[0]
                            stop_undo_on_error = cmd.stop_undo_on_error if cmd.stop_undo_on_error is not None \
                                else self.stop_undo_on_error
                            if r is False and cmd.stop_undo_on_error is True:
                                stop = True
                    duration = time.time() - start
                    left = max(timeout - duration, 0) if timeout else None

                if stop is True or (timeout is not None and (time.time() - start) > timeout):
                    break
            level -= 1
        return all(res) if len(res) > 0 else None

    @property
    def result(self) -> t.Dict[Id, CompletedProcess]:
        e = {}
        for n in self._dag.nodes:
            if n.success is not None:
                e.update(n.result)
        return e

    def __len__(self):
        return len(self._dag)

    def __iter__(self):
        return self._dag


class ProxyMixin(object):

    def __init__(self, server: Id, timeout: t.Union[int, float] = 300):
        """
        Parameters
        ----------
        server:
            Server Id to execute the command
        implementation:
            operation to perform.
        undo_implementation:
            undo operation to perform
        var_context:
            var_context to pass to the execution
        undo_on_error:
            if the invoke ended up with an error, the undo process will be executed if undo_on_error is True
        id_:
            command identifier
        timeout:
            timeout when waiting response from remote server when invoke and undo executed
        """
        self.__dict__['_server'] = server
        self.__dict__['_app'] = current_app._get_current_object()
        self.__dict__['_completion_event'] = threading.Event()
        self.__dict__['timeout'] = timeout
        self.__dict__['_command'] = None

    #
    # proxying (special cases)
    #

    def __getattr__(self, item):
        return getattr(object.__getattribute__(self, "_command"), item)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_command"), name)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_command"), name, value)

    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_command"))

    def __str__(self):
        return str(object.__getattribute__(self, "_command"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_command"))

    def __hash__(self):
        return hash(object.__getattribute__(self, "_command"))

    @property
    def server(self) -> 'Server':
        return self._server

    def callback_completion_event(self, event: Event):
        """callback executed on response to the invoke command on remote server
        """
        if 'success' in event.data:
            self._command._cp = CompletedProcess(success=event.data.get('success'),
                                                 stdout=event.data.get('stdout'),
                                                 stderr=event.data.get('stderr'),
                                                 rc=event.data.get('rc'),
                                                 start_time=datetime.datetime.strptime(event.data.get('start_time'),
                                                                                       defaults.DATETIME_FORMAT),
                                                 end_time=datetime.datetime.strptime(event.data.get('end_time'),
                                                                                     defaults.DATETIME_FORMAT))
            var_context = pickle.loads(base64.b64decode(event.data['var_context'].encode()))
            self._command.var_context.update_variables(var_context.extract_variables())
        else:
            self._command._cp = CompletedProcess(success=False, stdout=str(event.data),
                                                 stderr=f'Unknown message got on completion event.')

        self._completion_event.set()

    def invoke(self, timeout=None) -> bool:
        """
        invokes the command on the remote server

        Returns
        -------
        bool:
            True if command ended up gracefully. False otherwise

        Raises
        ------
        TimeoutError:
            raised when timeout reached while waiting the response back from the remote server
        """
        ctx = None
        if not has_app_context():
            ctx = self._app.app_context()
            ctx.push()
        try:
            auth = HTTPBearerAuth(
                create_access_token(self._command.var_context.globals['executor_id'], datetime.timedelta(seconds=15)))
            self.__dict__['_server'] = Server.query.get(self._server)
            timeout = timeout or self.timeout
            if self._command.success is None:
                try:
                    self.pre_processing(local=dict(vs=self.var_context, cp=self._cp))
                except Exception as e:
                    self._cp.pre_post_error = e
                    self._cp.success = False
                    return self._cp.success

                data = dict(operation=base64.b64encode(pickle.dumps(self._command.implementation)).decode('ascii'),
                            var_context=base64.b64encode(pickle.dumps(self._command.var_context)).decode('ascii'),
                            timeout=timeout,
                            step_id=str(self.id[1]),
                            execution=self._command.register.execution)
                resp = post(server=self.server, view_or_url='api_1_0.launch_operation', json=data, auth=auth)
                if resp.code == 202:
                    current_app.events.register(resp.msg.get('execution_id'), self.callback_completion_event)
                    event = self._completion_event.wait(timeout=timeout)
                    if event is not True:
                        self._command._cp = CompletedProcess(success=False, stdout='',
                                                             stderr=f'Timeout of {timeout} reached waiting '
                                                                    f'server operation completion')

                elif resp.code == 200:
                    self.callback_completion_event(Event(None, data=resp.msg))
                else:
                    self._command._cp = CompletedProcess(success=False, stdout='',
                                                         stderr=resp.msg, rc=resp.code)

                try:
                    self.post_processing(local=dict(vc=self._command.var_context, cp=self._command._cp))
                except Exception as e:
                    self._cp.pre_post_error = e
                    self._cp.success = False

                if self.register:
                    self.register.register_step_execution(self)
        finally:
            if ctx:
                ctx.pop()
        return self.success


class ProxyCommand(ProxyMixin, Command):

    def __init__(self, server_id: Id, *args, timeout: t.Union[int, float] = 300, **kwargs):
        ProxyMixin.__init__(self, server_id, timeout=timeout)
        object.__setattr__(self, "_command", Command(*args, **kwargs))


class ProxyUndoCommand(ProxyMixin, UndoCommand):

    def __init__(self, server_id: Id, *args, timeout: t.Union[int, float] = 300, **kwargs):
        ProxyMixin.__init__(self, server_id, timeout=timeout)
        object.__setattr__(self, "_command", UndoCommand(*args, **kwargs))


def _create_server_undo_command(executor, var_context, current_server, server_id: Id, s: Step, register, d=None,
                                s2cc=None) -> t.Optional[t.Union[UndoCommand, CompositeCommand]]:
    def iterate_tree(_cls, _step: Step, _d, _s2cc):
        if _step in _s2cc:
            _uc = _s2cc[_step]
        else:
            stop_on_error = _step.step_stop_on_error if _step.step_stop_on_error is not None else s.stop_undo_on_error
            _uc = _cls(create_operation(_step),
                       var_context=var_context.create_new_ctx(
                           defaults=ChainMap({**_step.parameters, 'server_id': server_id},
                                             _step.orchestration.parameters)),
                       id_=(str(server_id), str(_step.id)), post_process=_step.post_process,
                       stop_on_error=stop_on_error, register=register)
            _s2cc[_step] = _uc
        if _uc not in _d:
            _d[_uc] = []
        for child_step in _step.children_undo_steps:
            _d[_uc].append(iterate_tree(_cls, child_step, _d, _s2cc))
        return _uc

    d = d or {}
    s2cc = s2cc or {}
    if str(server_id) == str(current_server.id):
        u_cmd_cls = UndoCommand
    else:
        u_cmd_cls = partial(ProxyUndoCommand, server_id)
    uc = None
    for child_undo_step in s.children_undo_steps:
        uc = iterate_tree(u_cmd_cls, child_undo_step, d, s2cc)
    if len(d) == 1:
        return uc
    elif len(d) > 1:
        return CompositeCommand(dict_tree=d, stop_on_error=s.stop_undo_on_error,
                                stop_undo_on_error=None,
                                id_=('undo', str(s.id)),
                                executor=executor)


def create_cmd_from_orchestration(orchestration: Orchestration, var_context: VarContext,
                                  hosts: t.Dict[str, t.List[Id]],
                                  executor: concurrent.futures.Executor,
                                  register: 'RegisterStepExecution') -> CompositeCommand:
    current_server = Server.get_current()

    def create_do_cmd_from_step(_step: Step, _s2cc):
        d = {}
        if _step in _s2cc:
            cc = _s2cc[_step]
        else:
            for target in _step.target:
                for server_id in hosts[target]:
                    if str(server_id) == str(current_server.id):
                        cls = Command
                    else:
                        cls = partial(ProxyCommand, server_id)

                    c = cls(create_operation(_step),
                            undo_command=_create_server_undo_command(executor, var_context, current_server, server_id,
                                                                     _step, register),
                            var_context=var_context.create_new_ctx(
                                defaults=ChainMap({**_step.parameters, 'server_id': server_id},
                                                  _step.orchestration.parameters)),
                            stop_on_error=_step.stop_on_error,
                            stop_undo_on_error=None,
                            undo_on_error=_step.undo_on_error,
                            post_process=_step.post_process,
                            register=register,
                            id_=(str(server_id), str(_step.id)))

                    d[c] = []
            if len(d) == 1:
                cc = c
            elif len(d) > 1:
                cc = CompositeCommand(dict_tree=d,
                                      stop_on_error=False,
                                      stop_undo_on_error=False,
                                      id_=str(_step.id), executor=executor)
            else:
                cc = None
        return cc

    def iterate_tree(_step: Step, _d, _s2cc):
        if _step in s2cc_map:
            _c = _s2cc[_step]
        else:
            _c = create_do_cmd_from_step(_step, _s2cc)
            _s2cc[_step] = _c
            if _c not in _d:
                _d[_c] = []
            for child_step in _step.children_do_steps:
                _d[_c].append(iterate_tree(child_step, _d, _s2cc))
        return _c

    root_steps = orchestration.root
    tree = {}
    s2cc_map = {}
    for step in root_steps:
        iterate_tree(step, tree, s2cc_map)

    return CompositeCommand(dict_tree=tree,
                            stop_undo_on_error=orchestration.stop_undo_on_error,
                            stop_on_error=orchestration.stop_on_error,
                            id_=str(orchestration.id), executor=executor)


class RegisterStepExecution:
    def __init__(self, execution: OrchExecution, app: Flask = None):
        if app:
            self.app = app
        else:
            self.app = current_app._get_current_object()
        self.execution = execution.to_json()
        engine = db.get_engine()
        Session = sessionmaker(bind=engine)
        self.s = Session()

    def register_step_execution(self, command: ImplementationCommand):
        ctx = None
        if not has_app_context():
            ctx = self.app.app_context()
            ctx.push()

        try:
            e = StepExecution(orch_execution_id=self.execution['id'])
            e.load_completed_result(command._cp)
            e.params = dict(command.var_context)
            e.step_id = command.id[1]
            e.server_id = command.id[0]
            self.s.add(e)
            self.s.commit()
        finally:
            if ctx:
                ctx.pop()

    def __del__(self):
        self.s.close()



import collections
from threading import *
from typing import Optional, Callable, Any, Iterable, Mapping

from flask import current_app

from dm.web import db


def merge_iter(arg, c=None):
    c = c or {}
    if id(arg) in c:
        return c[id(arg)]
    elif isinstance(arg, db.Model):
        c.update({id(arg): db.session.merge(arg, load=False)})
        return c[id(arg)]
    elif isinstance(arg, tuple):
        t = tuple(merge_iter(item, c) for item in arg)
    elif isinstance(arg, list):
        for i in range(len(arg)):
            arg[i] = merge_iter(arg[i], c)
        return arg
    elif isinstance(arg, collections.MutableMapping):
        new_map = arg.__class__()
        for k, v in arg:
            new_map[merge_iter(k, c)] = merge_iter(v, c)
        return new_map
    else:
        return arg


class FlaskThread(Thread):
    def __init__(self, group: None = ..., target: Optional[Callable[..., Any]] = ..., name: Optional[str] = ...,
                 args: Iterable[Any] = ..., kwargs: Mapping[str, Any] = ..., *, daemon: Optional[bool] = ...,
                 reattach=True) -> None:
        self.app = current_app._get_current_object()
        self.reattach = reattach
        if daemon is None:
            daemon = True
        super().__init__(group, target, name, args, kwargs, daemon=daemon)

    def run(self) -> None:
        with self.app.app_context():
            if self.reattach:
                self._args = merge_iter(self._args)
                self._kwargs = merge_iter(self._kwargs)
            super().run()



import logging
import shlex
import typing as t
from typing import Iterable

from prompt_toolkit import HTML
from prompt_toolkit.completion import Completer, CompleteEvent, Completion, WordCompleter
from prompt_toolkit.completion.nested import NestedDict
from prompt_toolkit.document import Document

import dm.dshell.network as ntwrk
from dm.dshell.argparse_raise import GuessArgumentParser, create_parser, DictAction
from dm.dshell.utils import get_raw_text
from dm.utils.helpers import is_iterable_not_string


class DshellWordCompleter(WordCompleter):
    def get_completions(
            self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        # Get list of words.
        words = self.words
        if callable(words):
            words = words()

        # Get word/text before cursor.
        if self.sentence:
            word_before_cursor = document.text_before_cursor
        else:
            word_before_cursor = document.get_word_before_cursor(
                WORD=True, pattern=self.pattern
            )

        if self.ignore_case:
            word_before_cursor = word_before_cursor.lower()

        def word_matches(word: str, meta_data=None) -> bool:
            """ True when the word before the cursor matches. """
            if self.ignore_case:
                word = word.lower()
            raw = get_raw_text(meta_data)
            if self.match_middle:
                return (word_before_cursor in word and word_before_cursor != word) \
                       or raw is not None and word_before_cursor in raw
            else:
                return (word.startswith(
                    word_before_cursor) and word_before_cursor != word) or raw is not None and raw.startswith(
                    word_before_cursor)

        for a in words:
            display_meta = self.meta_dict.get(a, None)
            if word_matches(a, display_meta):
                if " " in a:
                    aa = a.replace("\"", '\\"')
                    yield Completion(f'"{aa}"', display=a, display_meta=display_meta)
                else:
                    yield Completion(a, -len(word_before_cursor), display_meta=display_meta)


class PathCompleter(Completer):

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        if document.char_before_cursor == ' ':
            json_data = dict()
            ntwrk.post('api_1_0.launch_command', )
        path = document.get_word_before_cursor(WORD=True)


class ResourceCompleter(Completer):

    def __init__(self, resource, key='id', meta_key=None, meta_html_format=None, ignore_case: bool = False,
                 match_middle: bool = True,
                 filters: t.List[t.Union[str, t.Tuple[str, str]]] = None,
                 transforms: t.Dict[str, t.Callable[[t.Any], t.Any]] = None,
                 resource_params: t.Dict['str', 'str'] = None) -> None:
        self.resource_params = resource_params or {}
        self.transforms = transforms
        self.match_middle = match_middle
        self.ignore_case = ignore_case
        self.resource = resource
        self.key = key
        self.meta_format = meta_html_format
        self.meta_key = meta_key
        self.filters = []
        for f in filters or []:
            if isinstance(f, str):
                self.filters.append((f, f.replace('-', '')))
            else:
                self.filters.append(f)

    def get_completions(self, document: Document, complete_event: CompleteEvent, var_filters: dict = None) -> \
            Iterable[Completion]:
        url_filters = {}
        for filter in self.filters:
            source, dest = filter
            if var_filters:
                value = var_filters.get(dest, None)
                if value:
                    if is_iterable_not_string(value):
                        value = ','.join(value)
                    url_filters.update({'filter[{dest}]': value})

        url = ntwrk.generate_url(self.resource, {**url_filters, **self.resource_params})
        res = ntwrk.request('get', url)
        if res.code == 200:
            words = []
            meta_words = {}
            for e in res.msg:
                if isinstance(e, dict) and e.get(self.key) not in words:
                    words.append(e.get(self.key))
                    if self.meta_key or self.meta_format:
                        if self.meta_format:
                            if self.transforms:
                                for key, transform in self.transforms.items():
                                    if key in e:
                                        val = e.get(key)
                                        try:
                                            e['key'] = transform(val)
                                        except:
                                            pass
                            try:
                                format = self.meta_format.format(**e)
                            except:
                                pass
                            else:
                                meta_words[e.get(self.key)] = HTML(format)
                        else:
                            meta_words[e.get(self.key)] = HTML(e.get(self.meta_key))
                elif isinstance(e, str):
                    words.append(e)

            completer = DshellWordCompleter(words, ignore_case=self.ignore_case,
                                            match_middle=self.match_middle, meta_dict=meta_words)
            for c in completer.get_completions(document, complete_event):
                yield c


class DshellCompleter(Completer):

    def __init__(self, options: t.Dict[str, t.Optional[t.Union[Completer, t.List[t.Dict]]]],
                 ignore_case: bool = True) -> None:

        self.options = options
        self.ignore_case = ignore_case

    @classmethod
    def from_nested_dict(cls, data: NestedDict) -> "DshellCompleter":
        """
        Create a `NestedCompleter`, starting from a nested dictionary data
        structure, like this:

        .. code::

            data = {
                'show': {
                    'version': None,
                    'interfaces': None,
                    'clock': None,
                    'ip': {'interface': {'brief'}}
                },
                'exit': None
                'enable': None
            }

        The value should be `None` if there is no further completion at some
        point. If all values in the dictionary are None, it is also possible to
        use a set instead.

        Values in this data structure can be a completers as well.
        """
        options: t.Dict[str, t.Optional[Completer]] = {}
        for key, value in data.items():
            if isinstance(value, dict):
                options[key] = cls.from_nested_dict(value)
            else:
                options[key] = value
        return cls(options)

    def get_completions(
            self, document: Document, complete_event: CompleteEvent
    ) -> t.Iterable[Completion]:

        # TODO: Problem with completing positionals. Solve argument resolution to know in which positional.

        # Split document.
        text = document.text_before_cursor.lstrip()
        stripped_len = len(document.text_before_cursor) - len(text)

        if text.endswith('-h') or text.endswith('--help'):
            return
        # If there is a space, check for the first term, and use a
        # subcompleter.
        if " " in text:
            first_term = text.split()[0]
            completer = self.options.get(first_term)

            # If we have a sub completer, use this for the completions.
            if isinstance(completer, Completer):
                remaining_text = text[len(first_term):].lstrip()
                move_cursor = len(text) - len(remaining_text) + stripped_len

                new_document = Document(
                    remaining_text,
                    cursor_position=document.cursor_position - move_cursor,
                )

                for c in completer.get_completions(new_document, complete_event):
                    yield c

            # we reached the bottom subcommand. Parse to see if we have to autocomplete an argument or its value
            else:
                options = {}
                params = {}
                dest_args = {}

                if not completer:
                    return
                for arg in completer:
                    if isinstance(arg, dict):
                        if arg.get('argument').startswith('-'):
                            options.update({arg.get('argument'): arg})
                        else:
                            params.update({arg.get('argument'): arg})

                        if 'dest' in arg:
                            dest_args.update({arg.get('dest'): arg})
                        else:
                            dest_args.update({arg.get('argument').lstrip('-'): arg})
                    if isinstance(arg, list):
                        for a in arg:
                            if a.get('argument').startswith('-'):
                                options.update({a.get('argument'): a})
                            else:
                                params.update({a.get('argument'): a})
                            if 'dest' in a:
                                dest_args.update({a.get('dest'): a})
                            else:
                                dest_args.update({a.get('argument').lstrip('-'): a})

                try:
                    words = shlex.split(text)
                except:
                    return
                if len(words) > 0 and words[-1] in options and text.endswith(words[-1]):
                    completer = DshellWordCompleter(words=list(options.keys()))
                    for c in completer.get_completions(document, complete_event):
                        yield c
                    for p in params:
                        if 'choices' in params[p]:
                            completer = DshellWordCompleter(params[p].get("choices"))
                        elif 'completer' in params[p]:
                            completer = params[p].get('completer')
                else:
                    parser = create_parser(completer, GuessArgumentParser(allow_abbrev=False))
                    finder = "F:I.N:D.E:R"
                    if document.char_before_cursor == ' ':
                        text = document.text + finder
                        current_word = finder
                    else:
                        text = document.text
                        current_word = document.get_word_before_cursor(WORD=True)

                    namespace = parser.parse_args(shlex.split(text)[1:])
                    values = dict(namespace._get_kwargs())

                    # find key related to current_word
                    for k, v in values.items():
                        if is_iterable_not_string(v):
                            if current_word in v:
                                break
                        else:
                            if v == current_word:
                                break
                    else:
                        k = None

                    # special case for DictAction
                    for dest, arg_def in dest_args.items():
                        if 'action' in arg_def and arg_def['action'] == DictAction and values[dest]:
                                for k, v in values[dest].items():
                                    # target
                                    if k == current_word:
                                        resp = ntwrk.get('api_1_0.orchestrationresource',
                                                         view_data={'orchestration_id': values['orchestration_id']})
                                        if resp.ok:
                                            needed_target = resp.msg['target']
                                            completer = DshellWordCompleter(
                                                [target for target in needed_target if target not in values[dest].keys()])
                                            for c in completer.get_completions(document, complete_event):
                                                yield c
                                            return

                                    elif current_word in v:
                                        completer = arg_def.get('completer', None)
                                        for c in completer.get_completions(document, complete_event):
                                            if c.text not in v:
                                                yield c
                                        if len(v) == 0 or len(v) == 1 and v[0] == finder:
                                            return
                                k = None
                    # get source     value
                    if k:
                        # completing an option argument value
                        if k in dest_args:
                            if 'choices' in dest_args[k]:
                                completer = DshellWordCompleter(dest_args[k].get("choices"))
                                for c in completer.get_completions(document, complete_event):
                                    yield c
                            completer = dest_args[k].get('completer', None)
                        else:
                            completer = None
                        # next argument can be a positional parameter or an optional argument
                        if k in params and current_word == finder:
                            # the user may want to add more options
                            _completer = DshellWordCompleter(words=list(options.keys()))
                            for c in _completer.get_completions(document, complete_event):
                                yield c
                        if completer:
                            if isinstance(completer, ResourceCompleter):
                                kwargs = dict(var_filters=values)
                            else:
                                kwargs = {}
                            for c in completer.get_completions(document, complete_event, **kwargs):
                                yield c
                    else:
                        completer = DshellWordCompleter(words=list(options.keys()))
                        for c in completer.get_completions(document, complete_event):
                            yield c
                        for p in params:
                            if getattr(namespace, p) is None:
                                if 'choices' in params[p]:
                                    completer = DshellWordCompleter(params[p].get("choices"))
                                elif 'completer' in params[p]:
                                    completer = params[p].get('completer')
                                for c in completer.get_completions(document, complete_event):
                                    yield c

        # No space in the input: behave exactly like `WordCompleter`.
        else:
            completer = DshellWordCompleter(
                list(self.options.keys()), ignore_case=self.ignore_case
            )
            for c in completer.get_completions(document, complete_event):
                yield c


server_name_completer = ResourceCompleter('api_1_0.serverlist', 'name')
server_completer = ResourceCompleter('api_1_0.serverlist', meta_key='name')
granule_completer = ResourceCompleter('api_1_0.granulelist')

orch_completer = ResourceCompleter('api_1_0.orchestrationlist', meta_html_format="<b>{name}</b>, ver <i>{version}</i>",
                                   filters=['--version'])
orch_name_completer = ResourceCompleter('api_1_0.orchestrationlist', 'name', filters=['--version'])
orch_ver_completer = ResourceCompleter('api_1_0.orchestrationlist', 'version', filters=['--name'])

action_completer = ResourceCompleter('api_1_0.actiontemplatelist',
                                     meta_html_format="<b>{name}</b>, ver <i>{version}</i>",
                                     filters=['--version'])
action_name_completer = ResourceCompleter('api_1_0.actiontemplatelist', 'name', filters=['--version'])
action_ver_completer = ResourceCompleter('api_1_0.actiontemplatelist', 'version', filters=['--name'])

software_completer = ResourceCompleter('api_1_0.softwarelist', meta_html_format="<b>{name}</b>, ver <i>{version}</i>", )
software_name_completer = ResourceCompleter('api_1_0.softwarelist', 'name', filters=['--version'])
software_ver_completer = ResourceCompleter('api_1_0.softwarelist', 'version', filters=['software'])
software_family_completer = ResourceCompleter('api_1_0.softwarelist', 'family')

logfed_completer = ResourceCompleter('api_1_0.loglist', resource_params={'params': 'human'},
                                     meta_html_format="<b>{target}</b>, "
                                                      "<i>{src_server}</i> -> "
                                                      "<i>{dst_server}</i>")

logger_completer = DshellWordCompleter([l for l in logging.root.manager.loggerDict.keys() if l.startswith('dshell')])



import copy
import ipaddress
import typing as t

from dm.domain.entities.base import UUIDistributedEntityMixin
from dm.utils.typos import UUID, IP as IPType
from dm.web import db

if t.TYPE_CHECKING:
    from dm.domain.entities import Server


class Gate(db.Model, UUIDistributedEntityMixin):
    __tablename__ = "D_gate"
    order = 20

    server_id = db.Column(UUID, db.ForeignKey('D_server.id'))
    dns = db.Column(db.String(100))
    ip = db.Column(IPType)
    port = db.Column(db.Integer, nullable=False)
    hidden = db.Column(db.Boolean, default=False)

    __table_args__ = (db.UniqueConstraint('server_id', 'ip', 'dns'),)

    server = db.relationship("Server", back_populates="gates")

    def __init__(self, server: 'Server', port: int, dns: str = None,
                 ip: t.Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address] = None, **kwargs):
        UUIDistributedEntityMixin.__init__(self, **kwargs)
        self.server = server
        self.port = port
        self.ip = ipaddress.ip_address(ip) if isinstance(ip, str) else ip
        self.dns = dns
        if not (self.dns or self.ip):
            self.dns = server.name
        self.hidden = kwargs.get('hidden', False) or False

    def __str__(self):
        return f'{self.dns or self.ip}:{self.port}'

    def to_json(self, human=False):
        data = super().to_json()
        data.update(server_id=str(self.server.id) if self.server.id else None, ip=str(self.ip) if self.ip else None,
                    dns=self.dns, port=self.port, hidden=self.hidden)
        return data

    @classmethod
    def from_json(cls, kwargs):
        from dm.domain.entities import Server
        server = kwargs.pop('server', None)
        kwargs = copy.deepcopy(kwargs)
        if server:
            kwargs['server'] = server
        kwargs['ip'] = ipaddress.ip_address(kwargs.get('ip')) if isinstance(kwargs.get('ip'), str) else kwargs.get('ip')
        if 'server_id' in kwargs and kwargs['server_id'] is not None:
            kwargs['server'] = Server.query.filter_by(id=kwargs.get('server_id')).one()
            kwargs.pop('server_id')
        return super().from_json(kwargs)



"""
usage: dshell [--version] [--username USER] [--password PASSWORD|--token TOKEN] [--server SERVER] [--port PORT]
              [--config-file CONFIG_FILE] [COMMAND [ARGS...]]

options:
   -u, --username USER              Login user
   --password PASSWORD              Password used to login
   -t, --token TOKEN                Refresh token used to authenticate
   -s, --server SERVER              Server to communicate with
   -p, --port PORT                  Port to communicate with [default: 5000]
   -c, --config-file CONFIG_FILE    configuration file [default: ~/.dshell]

The most commonly used dshell commands are:
   status    Status of a node
   ping      Ping a node
   server    Server Resource
   orch      Orchestration Resource, run Orchestrations
   action    Action Resource
   exec      Execution Resource
   software  Software Resource, send software
   logfed    LogFed Resource

When no command specified dshell starts in interactive mode

See 'dshell <command> --help' for more information on a specific command.

"""
import configparser
import importlib
import logging
import os
import sys
from argparse import ArgumentParser
from os.path import expanduser

from docopt import docopt

import dm.dshell.network as ntwrk
from dm.dshell.argparse_raise import create_parser
from dm.dshell.commands import nested_dict
from dm.dshell.environ import set_dict_in_environ
from dm.dshell.interactive import interactive, call_func_with_signature

basename = os.path.dirname(os.path.abspath(__file__))

commands = 'cmd exec ping logfed server software status transfer'.split()
batch_commands = 'action orch'.split()  # commands that interact differently from interactive mode


def load_config_file(file=None):
    data = dict(USERNAME=None, TOKEN=None, SERVER=None, PORT=None)
    config = configparser.ConfigParser()
    if file and os.path.exists(file):
        config.read(file)
        if 'AUTH' in config:
            data['USERNAME'] = config['AUTH'].get('username', None)
            data['TOKEN'] = config['AUTH'].get('token', None)
        if 'REMOTE' in config:
            data['SERVER'] = config['REMOTE'].get('server', None)
            data['PORT'] = config['REMOTE'].get('port', None)

    return data


def main():
    ch = logging.StreamHandler()
    logger = logging.getLogger('dshell')
    logger.addHandler(ch)

    # parse args
    args = docopt(__doc__,
                  version=f'dshell version 1.0',
                  options_first=True)

    # load data from config file
    data = load_config_file(os.path.join(expanduser(args['--config-file'])))

    SERVER = args['--server'] or data['SERVER'] or os.environ.get('DM_SERVER', None)
    PORT = data['PORT'] or os.environ.get('DM_PORT', None)
    if '--port' in sys.argv or PORT is None:
        PORT = args['--port']

    set_dict_in_environ({'SERVER': SERVER, 'PORT': PORT, 'SCHEME': 'https', 'SSL_VERIFY': False,
                         'FILE_HISTORY': '~/.dshell_history'})

    # process args
    argv = [args['COMMAND']] + args['ARGS']
    if args['COMMAND'] in commands:
        ntwrk.bootstrap_auth(args['--username'] or data['USERNAME'], args['--password'],
                             args['--token'] or data['TOKEN'])
        parser = create_parser({args['COMMAND']: nested_dict[args['COMMAND']]}, parser=ArgumentParser(prog="dshell"))
        namespace = parser.parse_args(argv)

        if hasattr(namespace, 'func'):
            call_func_with_signature(dict(namespace._get_kwargs()))
    elif args['COMMAND'] in batch_commands:
        ntwrk.bootstrap_auth(args['--username'] or data['USERNAME'], args['--password'],
                             args['--token'] or data['TOKEN'])
        module = importlib.import_module('.dshell_%s' % args['COMMAND'], 'dm.dshell.batch')
        module.main(argv)
    elif args['COMMAND'] is None:
        try:
            ntwrk.bootstrap_auth(args['--username'] or data['USERNAME'], args['--password'],
                                 args['--token'] or data['TOKEN'])
        except Exception as e:
            print(str(e))
        interactive()
    else:
        exit("%r is not a dshell command. See 'dshell --help'." % args['COMMAND'])


if __name__ == '__main__':
    sys.exit(main())



from flask import request
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import Orchestration
from dm.web import db
from dm.web.decorators import securizer, forward_or_dispatch, validate_schema, lock_catalog
from dm.web.helpers import filter_query, check_param_in_uri
from dm.web.json_schemas import orchestration_post, orchestration_patch


class OrchestrationList(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self):
        query = filter_query(Orchestration, request.args).order_by(Orchestration.created_at)
        return [o.to_json(add_target=check_param_in_uri('target'), add_params=check_param_in_uri('vars'),
                          add_steps=check_param_in_uri('steps'), add_action=check_param_in_uri('action')) for o in
                query.all()]

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(orchestration_post)
    @lock_catalog
    def post(self):
        json_data = request.get_json()
        generated_version = False
        if 'version' not in json_data:
            generated_version = True
            json_data['version'] = Orchestration.query.filter_by(name=json_data['name']).count() + 1
        o = Orchestration(**json_data)
        db.session.add(o)
        db.session.commit()
        resp_data = {'id': str(o.id)}
        if generated_version:
            resp_data.update(version=o.version)
        return resp_data, 201


class OrchestrationResource(Resource):
    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, orchestration_id):
        return Orchestration.query.get_or_404(orchestration_id).to_json(add_target=True, add_params=True)

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(orchestration_patch)
    @lock_catalog
    def patch(self, orchestration_id):
        o = Orchestration.query.get_or_404(orchestration_id)
        for k, v in request.get_json().items():
            setattr(o, k, v)
        db.session.commit()
        return {}, 204


"""
Exceptions for Domain Errors
IDs from 100 to 999
"""


class DomainException(Exception):
    """
    Base Class exception
    """
    _id = 100


class InjectionError(DomainException):
    """
    Exception raised when trying to add an id to Step that already exists
    """
    _id = 102


class CycleError(DomainException):
    """
    Exception raised when trying to add a node dependency which makes a cycle in the orchestration
    """
    _id = 103


# Locker Exceptions
class LockerError(DomainException):
    """Base Exception Class for Locker module"""
    _id = 104
    pass


class StateError(LockerError):
    """Exception for incorrect states"""
    _id = 105


class StateAlreadyInLock(StateError):
    """State already in LOCKED state"""
    _id = 106


class StateAlreadyInPreventingLock(StateError):
    """State already in PREVENTING LOCK state"""
    _id = 107


class StateAlreadyInUnlock(StateError):
    """State already in UNLOCK state"""
    _id = 108


class StateTransitionError(StateError):
    """State must be in PREVENTING LOCK to change to LOCK"""
    _id = 109

    def __init__(self, before, after):
        self.before = before
        self.after = after


class ApplicantError(LockerError):
    """Another applicant has the locker"""
    _id = 110


class PriorityError(LockerError):
    """A locker with higher priority is locked or trying to lock"""
    _id = 111


class CatalogError(DomainException):
    """
    Error raised when repository functions not set
    """
    _id = 120


class DataMarkError(CatalogError):
    """
    Error raised when past data mark generated
    """
    _id = 121


class NoDataMarkSet(CatalogError):
    """
    Error raised when catalog bypassed but no datemark passed into entity
    """
    _id = 122



from requests.auth import AuthBase


class HTTPBearerAuth(AuthBase):
    def __init__(self, token):
        self.token = token

    def __eq__(self, other):
        return self.token == getattr(other, 'token', None)

    def __ne__(self, other):
        return not self == other

    def __call__(self, r):
        if hasattr(r, 'headers'):
            r.headers.update(self.header)
        else:
            r['headers'].update(self.header)
            r.pop('auth', None)
        return r

    @property
    def header(self):
        return {'Authorization': str(self)}

    def __str__(self):
        return 'Bearer ' + self.token


import json
import typing as t
import uuid
from datetime import datetime

from dm import defaults
from dm.domain.entities.base import EntityReprMixin
from dm.utils.typos import UUID, UtcDateTime
from dm.web import db
from .server import Server
from ...utils.helpers import is_iterable_not_string, get_now

if t.TYPE_CHECKING:
    from dm.use_cases.operations import CompletedProcess


# class Status(enum.Enum):
#     PENDING = enum.auto()
#     EXECUTING = enum.auto()
#     FINISHED = enum.auto()
#     ROLLING_BACK = enum.auto()
#     ROLLED_BACK = enum.auto()
#     CANCELLED = enum.auto()
#     ERROR = enum.auto()


class StepExecution(db.Model, EntityReprMixin):
    __tablename__ = 'L_execution'

    id = db.Column(UUID, primary_key=True, default=uuid.uuid4)

    start_time = db.Column(UtcDateTime(timezone=True), nullable=False, default=get_now())
    end_time = db.Column(UtcDateTime(timezone=True))
    params = db.Column(db.JSON)
    rc = db.Column(db.Integer)
    stdout = db.Column(db.Text)
    stderr = db.Column(db.Text)
    success = db.Column(db.Boolean)
    step_id = db.Column(UUID, db.ForeignKey('D_step.id'), nullable=False)
    server_id = db.Column(UUID, db.ForeignKey('D_server.id'))
    orch_execution_id = db.Column(UUID)

    step = db.relationship("Step")
    server = db.relationship("Server", foreign_keys=[server_id])
    orch_execution = db.relationship("OrchExecution", foreign_keys=[orch_execution_id],
                                     primaryjoin="OrchExecution.id==StepExecution.orch_execution_id",
                                     uselist=False, backref="step_executions")

    def load_completed_result(self, cp: 'CompletedProcess'):
        self.success = cp.success
        self.stdout = cp.stdout
        self.stderr = cp.stderr
        if cp.pre_post_error:
            self.stderr += '\nError on pre/post process: '
            if str(cp.pre_post_error):
                self.stderr += str(cp.pre_post_error)
            else:
                self.stderr += str(cp.pre_post_error.__class__.__name__)
        self.rc = cp.rc
        self.start_time = cp.start_time
        self.end_time = cp.end_time

    def to_json(self, human=False):
        data = {}
        if self.id:
            data.update(id=str(self.id))
        if self.start_time:
            data.update(start_time=self.start_time.strftime(defaults.DATETIME_FORMAT))
        if self.end_time:
            data.update(end_time=self.end_time.strftime(defaults.DATETIME_FORMAT))
        data.update(params=self.params)
        data.update(step_id=str(getattr(self.step, 'id', None)) if getattr(self.step, 'id', None) else None)
        data.update(rc=self.rc, success=self.success)
        if human:
            data.update(server=str(self.server) if self.server else None)
            try:
                stdout = json.loads(self.stdout)
            except:
                stdout = self.stdout
            try:
                stderr = json.loads(self.stderr)
            except:
                stderr = self.stderr
        else:
            data.update(
                server_id=str(getattr(self.server, 'id', None)) if getattr(self.server, 'id', None) else None)
            stdout = self.stdout
            stderr = self.stderr
        data.update(stdout=stdout)
        data.update(stderr=stderr)
        return data


class OrchExecution(db.Model, EntityReprMixin):
    __tablename__ = 'L_orch_execution'

    id = db.Column(UUID, primary_key=True, default=uuid.uuid4)
    start_time = db.Column(UtcDateTime(timezone=True), nullable=False, default=get_now)
    end_time = db.Column(UtcDateTime(timezone=True))
    orchestration_id = db.Column(UUID, db.ForeignKey('D_orchestration.id'), nullable=False)
    target = db.Column(db.JSON)
    params = db.Column(db.JSON)
    executor_id = db.Column(UUID, db.ForeignKey('D_user.id'))
    service_id = db.Column(UUID, db.ForeignKey('D_service.id'))
    success = db.Column(db.Boolean)
    undo_success = db.Column(db.Boolean)
    message = db.Column(db.Text)
    parent_orch_execution_id = db.Column(UUID, db.ForeignKey('L_orch_execution.id'))
    server_id = db.Column(UUID, db.ForeignKey('D_server.id'))

    orchestration = db.relationship("Orchestration")
    _executor = db.relationship("User")
    service = db.relationship("Service")
    parent_orch_execution = db.relationship("OrchExecution", uselist=False)
    server = db.relationship("Server", foreign_keys=[server_id])

    @property
    def executor(self):
        return self._executor if self._executor else self.parent_orch_execution

    def to_json(self, add_step_exec=False, human=False):
        data = {}
        if self.id:
            data.update(id=str(self.id))
        if self.start_time:
            data.update(start_time=self.start_time.strftime(defaults.DATETIME_FORMAT))
        if self.end_time:
            data.update(end_time=self.end_time.strftime(defaults.DATETIME_FORMAT))
        if human:
            # convert target ids to server names
            d = {}
            for k, v in self.target.items():
                if is_iterable_not_string(v):
                    d[k] = [str(Server.query.get(s) or s) for s in v]
                else:
                    d[k] = str(Server.query.get(v) or v)
            data.update(target=d)
            data.update(executor=str(self._executor) if self._executor else None)
            data.update(service=str(self.service) if self.service else None)
            if self.orchestration:
                data.update(
                    orchestration=dict(id=str(self.orchestration.id), name=self.orchestration.name,
                                       version=self.orchestration.version))
            else:
                data.update(
                    orchestration=None)
            if self.server:
                data.update(server=dict(id=str(self.server.id), name=self.server.name))
            else:
                data.update(server=None)
        else:
            data.update(target=self.target)
            data.update(
                orchestration_id=str(getattr(self.orchestration, 'id', None)) if getattr(self.orchestration, 'id',
                                                                                         None) else None)
            data.update(
                executor_id=str(getattr(self._executor, 'id', None)) if getattr(self._executor, 'id', None) else None)
            data.update(
                service_id=str(getattr(self.service, 'id', None)) if getattr(self.service, 'id', None) else None)
            data.update(
                server_id=str(getattr(self.server, 'id', None)) if getattr(self.server,
                                                                           'id', None) else None)
        data.update(params=self.params)
        data.update(success=self.success)
        data.update(undo_success=self.undo_success)
        data.update(message=self.message)
        if self.parent_orch_execution:
            data.update(parent_orch_execution_id=str(self.parent_orch_execution.id))
        if add_step_exec:
            data.update(steps=[e.to_json(human) for e in self.step_executions])
        return data

    @classmethod
    def from_json(cls, kwargs):
        if 'start_time' in kwargs:
            kwargs['start_time'] = datetime.strptime(kwargs.get('start_time'), defaults.DATETIME_FORMAT)
        if 'end_time' in kwargs:
            kwargs['end_time'] = datetime.strptime(kwargs.get('end_time'), defaults.DATETIME_FORMAT)
        try:
            o = cls.query.get(kwargs.get('id'))
        except RuntimeError as e:
            o = None
        if o:
            for k, v in kwargs.items():
                if getattr(o, k) != v:
                    setattr(o, k, v)
            return o
        else:
            return cls(**kwargs)

